from typing import Callable, Optional

import torch


def lif_forward_single(
    input_data: torch.Tensor,
    alpha_mem: float,
    alpha_syn: float,
    state: dict,
    spike_threshold: float,
    spike_fn: Callable,
    reset_fn: Callable,
    surrogate_grad_fn: Callable,
    min_v_mem: Optional[float],
    norm_input: bool,
):
    # if t_syn was provided, we're going to use synaptic current dynamics
    if alpha_syn is not None:
        state["i_syn"] = alpha_syn * (state["i_syn"] + input_data)
    else:
        state["i_syn"] = input_data

    if norm_input:
        synaptic_input = (1 - alpha_mem) * state["i_syn"]
    else:
        synaptic_input = state["i_syn"]

    # Decay the membrane potential and add the input currents which are normalised by tau
    state["v_mem"] = alpha_mem * state["v_mem"] + synaptic_input

    # generate spikes and adjust v_mem
    if spike_fn:
        input_tensors = [state[name] for name in spike_fn.required_states]
        spikes = spike_fn.apply(*input_tensors, spike_threshold, surrogate_grad_fn)
        state = reset_fn(spikes, state, spike_threshold)
    else:
        spikes = state["v_mem"].clone()
        state = state.copy()

    if min_v_mem is not None:
        state["v_mem"] = (
            torch.nn.functional.relu(state["v_mem"] - min_v_mem) + min_v_mem
        )
    return spikes, state


def lif_forward(
    input_data: torch.Tensor,
    alpha_mem: float,
    alpha_syn: float,
    state: dict,
    spike_threshold: float,
    spike_fn: Callable,
    reset_fn: Callable,
    surrogate_grad_fn: Callable,
    min_v_mem: float,
    norm_input: bool,
    record_states: bool = False,
):
    n_time_steps = input_data.shape[1]
    state_names = list(state.keys())

    output_spikes = []
    if record_states:
        recordings = {name: [] for name in state_names}

    for step in range(n_time_steps):
        spikes, state = lif_forward_single(
            input_data=input_data[:, step],
            alpha_mem=alpha_mem,
            alpha_syn=alpha_syn,
            state=state,
            spike_threshold=spike_threshold,
            spike_fn=spike_fn,
            reset_fn=reset_fn,
            surrogate_grad_fn=surrogate_grad_fn,
            min_v_mem=min_v_mem,
            norm_input=norm_input,
        )
        output_spikes.append(spikes)
        if record_states:
            for name in state_names:
                recordings[name].append(state[name].clone())

    if record_states:
        record_dict = {name: torch.stack(vals, 1) for name, vals in recordings.items()}
    else:
        record_dict = dict()

    return torch.stack(output_spikes, 1), state, record_dict


def lif_recurrent(
    input_data: torch.Tensor,
    alpha_mem: float,
    alpha_syn: float,
    state: dict,
    spike_threshold: float,
    spike_fn: Callable,
    reset_fn: Callable,
    surrogate_grad_fn: Callable,
    min_v_mem: Optional[float],
    norm_input: bool,
    rec_connect: torch.nn.Module,
    record_states: bool = False,
):
    batch_size, n_time_steps, *trailing_dim = input_data.shape
    state_names = list(state.keys())

    output_spikes = []
    recordings = []
    rec_out = torch.zeros((batch_size, *trailing_dim), device=input_data.device)
    for step in range(n_time_steps):
        total_input = input_data[:, step] + rec_out

        spikes, state = lif_forward_single(
            input_data=total_input,
            alpha_mem=alpha_mem,
            alpha_syn=alpha_syn,
            state=state,
            spike_threshold=spike_threshold,
            spike_fn=spike_fn,
            reset_fn=reset_fn,
            surrogate_grad_fn=surrogate_grad_fn,
            min_v_mem=min_v_mem,
            norm_input=norm_input,
        )
        output_spikes.append(spikes)
        if record_states:
            recordings.append(state)

        # compute recurrent output that will be added to the input at the next time step
        rec_out = rec_connect(spikes).reshape((batch_size, *trailing_dim))

    record_dict = {}
    if record_states:
        for state_name in state_names:
            record_dict[state_name] = torch.stack(
                [item[state_name] for item in recordings], 1
            )
    return torch.stack(output_spikes, 1), state, record_dict



"""
==============================
Leaky Integrate and Fire (LIF)
==============================
The :class:`~sinabs.layers.LIF` layer. This neuron integrates the input and decays its state at every time step.
"""

import torch
from utils import plot_evolution

import sinabs.layers as sl

const_current = torch.ones((1, 100, 1)) * 0.03
single_current = torch.zeros((1, 100, 1))
single_current[:, 0] = 0.1

lif_neuron = sl.LIF(tau_mem=40.0, norm_input=False, record_states=True)
plot_evolution(lif_neuron, const_current)


# By default, no synaptic dynamics are used. We can enable that by setting tau_syn. Note that instead of a constant current, we now provide input only at the first time step.

lif_neuron = sl.LIF(tau_mem=40.0, tau_syn=30.0, norm_input=False, record_states=True)
plot_evolution(lif_neuron, single_current)



import torch.nn as nn

from sinabs.activation import Quantize


class QuantizeLayer(nn.Module):
    """Layer that quantizes the input, i.e. returns floor(input).

    Parameters:
        quantize: If False, this layer will pass on the input without modifying it.
    """

    def __init__(self, quantize=True):
        super().__init__()
        self.quantize = quantize

    def forward(self, data):
        if self.quantize:
            return Quantize.apply(data)
        else:
            return data



import sinabs

project = "Sinabs"
copyright = "2019-present, SynSense"
author = "employees of SynSense"

master_doc = "index"

extensions = [
    "myst_nb",
    "pbr.sphinxext",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx_gallery.gen_gallery",
]

autodoc_typehints = "both"

sphinx_gallery_conf = {
    "examples_dirs": "gallery/",  # path to your example scripts
    "gallery_dirs": "auto_examples",  # path to where to save gallery generated output
    # "backreferences_dir": "gen_modules/backreferences",
    "doc_module": ("sinabs",),
    "download_all_examples": False,
    "ignore_pattern": r"utils\.py",
}

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True

# MyST settings
suppress_warnings = ["myst.header"]
nb_execution_timeout = 300
nb_execution_excludepatterns = [
    "LeNet_5_EngChinese.ipynb",
    "bptt.ipynb",
    "weight_transfer_mnist.ipynb",
    "synops_loss_ann.ipynb",
]
# nb_execution_mode = "off"

templates_path = ["_templates"]

exclude_patterns = ["_build", "**.ipynb_checkpoints"]

html_title = sinabs.__version__
html_theme = "sphinx_book_theme"
html_logo = "_static/sinabs-logo-lowercase-whitebg.png"
html_show_sourcelink = True
html_sourcelink_suffix = ""

html_static_path = ["_static"]

html_theme_options = {
    "repository_url": "https://github.com/synsense/sinabs",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
    "repository_branch": "develop",
    "path_to_docs": "docs",
    "use_fullscreen_button": True,
}



import warnings
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn

from .layers import StatefulLayer
from .synopcounter import SNNAnalyzer
from .utils import get_activations, get_network_activations

ArrayLike = Union[np.ndarray, List, Tuple]


class Network(torch.nn.Module):
    """Class of a spiking neural network.

    Attributes:
        spiking_model: torch.nn.Module, a spiking neural network model
        analog_model: torch.nn.Module, an artifical neural network model
        input_shape: Tuple, size of input
        synops: If True (default: False), register hooks for counting synaptic \
        operations during forward passes, instantiating `sinabs.SNNSynOpCounter`.
    """

    def __init__(
        self,
        analog_model=None,
        spiking_model=None,
        input_shape: Optional[ArrayLike] = None,
        synops: bool = False,
        batch_size: int = 1,
        num_timesteps: int = 1,
    ):
        super().__init__()
        self.spiking_model: nn.Module = spiking_model
        self.analog_model: nn.Module = analog_model
        self.input_shape = input_shape

        self.synops = synops
        if synops:
            self.synops_counter = SNNAnalyzer(self.spiking_model)

        if input_shape is not None and spiking_model is not None:
            self._compute_shapes(
                input_shape, batch_size=batch_size, num_timesteps=num_timesteps
            )

    @property
    def layers(self):
        return list(self.spiking_model.named_children())

    def _compute_shapes(self, input_shape, batch_size=1, num_timesteps=1):
        def hook(module, inp, out):
            module.out_shape = out.shape[1:]

        hook_list = []
        for layer in self.spiking_model.modules():
            this_hook = layer.register_forward_hook(hook)
            hook_list.append(this_hook)

        device = next(self.parameters()).device

        # Infer shape
        if batch_size is None:
            batch_size = 1
        if num_timesteps is None:
            num_timesteps = 1
        shape = [batch_size * num_timesteps] + list(input_shape)
        dummy_input = torch.zeros(shape, requires_grad=False).to(device)
        # do a forward pass
        self(dummy_input)

        [this_hook.remove() for this_hook in hook_list]

    def forward(self, tsrInput) -> torch.Tensor:
        """Forward pass for this model."""
        return self.spiking_model(tsrInput)

    def compare_activations(
        self,
        data,
        name_list: Optional[ArrayLike] = None,
        compute_rate: bool = False,
        verbose: bool = False,
    ) -> Tuple[np.ndarray, np.ndarray, str]:
        """Compare activations of the analog model and the SNN for a given data sample.

        Args:
            data (np.ndarray):      Data to process
            name_list (List[str]):  list of all layer names (str) whose activations need to be compared
            compute_rate (bool):    True if you want to compute firing rate. By default spike count is returned
            verbose (bool):         print debugging logs to the terminal
        Returns:
            tuple: A tuple of lists (ann_activity, snn_activity, name_list)
                - ann_activity: output activity of the ann layers
                - snn_activity: output activity of the snn layers
                - name_list: spiking layers' name list for plotting comparison
        """
        if name_list is None:
            name_list = ["Input"]
            for layer_name, lyr in self.spiking_model.named_modules():
                if isinstance(lyr, StatefulLayer):
                    name_list.append(layer_name)

        if verbose:
            print("Comparing activations for {0}".format(name_list))

        # Calculate activations for the torch analog model
        if compute_rate:
            tsrAnalogData = data.mean(0).unsqueeze(0)
        else:
            tsrAnalogData = data.sum(0).unsqueeze(0)
        with torch.no_grad():
            analog_activations = get_activations(
                self.analog_model, tsrAnalogData, name_list=name_list
            )

        # Calculate activations for spiking model
        spike_rates = get_network_activations(
            self.spiking_model, data, name_list=name_list, bRate=compute_rate
        )
        return analog_activations, spike_rates, name_list

    def plot_comparison(
        self, data, name_list: Optional[ArrayLike] = None, compute_rate=False
    ):
        """Plots a scatter plot of all the activations.

        Args:
            data: Data to be processed
            name_list: ArrayLike with names of all the layers of interest to be compared
            compute_rate: Compare firing rates instead of spike count
        Returns:
            tuple: A tuple of lists (ann_activity, snn_activity)
                - ann_activity: output activity of the ann layers
                - snn_activity: output activity of the snn layers
        """
        import pylab

        analog_activations, spike_rates, name_list = self.compare_activations(
            data, name_list=name_list, compute_rate=compute_rate
        )
        for nLyrIdx in range(len(name_list)):
            pylab.scatter(
                spike_rates[nLyrIdx],
                analog_activations[nLyrIdx],
                label=name_list[nLyrIdx],
            )
        if compute_rate:
            pylab.xlabel("Spike rates (Hz)")
        else:
            pylab.xlabel("# Spike count")
        pylab.ylabel("Analog activations")
        pylab.legend()
        return analog_activations, spike_rates

    def reset_states(
        self,
        randomize: bool = False,
        value_ranges: Optional[List[Dict[str, Tuple[float, float]]]] = None,
    ):
        """Reset all neuron states in the submodules.

        Parameters
        ----------
        randomize: Bool
            If true, reset the states between a range provided. Else, the states are reset to zero.
        value_ranges: Optional[List[Dict[str, Tuple[float, float]]]]
            A list of value_range dictionaries with the same length as the total stateful layers in the module.
            Each dictionary is a key value pair: buffer_name -> (min, max) for each state that needs to be reset.
            The states are reset with a uniform distribution between the min and max values specified.
            Any state with an undefined key in this dictionary will be reset between 0 and 1
            This parameter is only used if randomize is set to true.
        """

        if value_ranges:
            num_stateful_layers = len(
                [None for mod in self.modules() if isinstance(mod, StatefulLayer)]
            )
            if len(value_ranges) != num_stateful_layers:
                raise TypeError(
                    "The number of entries in value_ranges does not match the number of stateful sub modules"
                )
        i = 0
        for lyr in self.modules():
            if isinstance(lyr, StatefulLayer):
                if value_ranges is None:
                    vr = None
                else:
                    vr = value_ranges[i]
                    i += 1
                lyr.reset_states(randomize=randomize, value_ranges=vr)

    def zero_grad(self, set_to_none: bool = False) -> None:
        for lyr in self.spiking_model:
            lyr.zero_grad(set_to_none)

    def get_synops(self, num_evs_in=None) -> dict:
        """Please see docs for `sinabs.SNNSynOpCounter.get_synops()`."""
        if num_evs_in is not None:
            warnings.warn("num_evs_in is deprecated and has no effect")

        return self.synops_counter.get_synops()


def get_parent_module_by_name(
    root: torch.nn.Module, name: str
) -> Tuple[torch.nn.Module, str]:
    """Find a nested Module of a given name inside a Module, and return its parent Module.

    Args:
        root: The Module inside which to look for the nested Module
        name: Name of the Module that is being searched for within root. Must
              contain all parent modules, separated by a `.` , e.g.
              "root.nested_module1.nested_module2.desired_module"
    Returns:
        torch.nn.Module: The Module that contains the Module with the given name. In the example
        above this would be `nested_module2`.
        str: The name of the child, without parent modules, e.g. "desired_module"
    """
    if "." not in name:
        if not hasattr(root, name):
            raise KeyError(f"The requested module `{name}` could not be found.")

        return root, name
    else:
        child_name, *rest = name.split(".")
        child = getattr(root, child_name)
        return get_parent_module_by_name(child, ".".join(rest))


def infer_module_device(module: torch.nn.Module) -> Union[torch.device, None]:
    """Infere on which device a module is operating by first looking at its parameters and then, if
    no parameters are found, at its buffers.

    Args:
        module: The module whose device is to be inferred.

    Returns:
        torch.device: The device of 'module', or `None` if no device has been found.
    """

    try:
        return next(module.parameters()).device
    except StopIteration:
        # No parameters, try buffers
        try:
            return next(module.buffers()).device
        except StopIteration:
            # No buffers, don't infer device
            return None



import torch.autograd


class Quantize(torch.autograd.Function):
    """PyTorch-compatible function that applies a floor() operation on the input, while providing a
    surrogate gradient (equivalent to that of a linear function) in the backward pass."""

    @staticmethod
    def forward(ctx, inp):
        """"""
        return inp.floor()

    @staticmethod
    def backward(ctx, grad_output):
        """"""
        grad_input = grad_output.clone()
        return grad_input


class StochasticRounding(torch.autograd.Function):
    """PyTorch-compatible function that applies stochastic rounding. The input x.

    is quantized to ceil(x) with probability (x - floor(x)), and to floor(x)
    otherwise. The backward pass is provided as a surrogate gradient
    (equivalent to that of a linear function).
    """

    @staticmethod
    def forward(ctx, inp):
        """"""
        int_val = inp.floor()
        frac = inp - int_val
        output = int_val + (torch.rand_like(inp) < frac).float()
        return output

    @staticmethod
    def backward(ctx, grad_output):
        """"""
        grad_input = grad_output.clone()
        return grad_input



import torch

from sinabs.activation import Quantize, StochasticRounding


class NeuromorphicReLU(torch.nn.Module):
    """NeuromorphicReLU layer. This layer is NOT used for Sinabs networks; it's useful while
    training analogue pyTorch networks for future use with Sinabs.

    Parameters:
        quantize: Whether or not to quantize the output (i.e. floor it to \
        the integer below), in order to mimic spiking behavior.
        fanout: Useful when computing the number of SynOps of a quantized \
        NeuromorphicReLU. The activity can be accessed through \
        NeuromorphicReLU.activity, and is multiplied by the value of fanout.
        stochastic_rounding: Upon quantization, should the value be rounded stochastically or floored
        Only done during training. During evaluation mode, the value is simply floored
    """

    def __init__(self, quantize=True, fanout=1, stochastic_rounding=False):
        super().__init__()
        self.quantize = quantize
        self.stochastic_rounding = stochastic_rounding
        self.fanout = fanout

    def forward(self, inp):
        output = torch.nn.functional.relu(inp)
        if self.quantize:
            if self.stochastic_rounding:
                if self.training:
                    output = StochasticRounding.apply(output)
                else:
                    output = output.round()
            else:
                output = Quantize.apply(output)

        self.activity = output.sum() / len(output) * self.fanout
        return output



import numpy as np
import torch
from torch import nn

import sinabs
import sinabs.layers as sl
from sinabs import SNNAnalyzer, SynOpCounter
from sinabs.layers import IAFSqueeze, NeuromorphicReLU


class Model(torch.nn.Sequential):
    def __init__(self):
        super().__init__(
            torch.nn.Conv2d(
                in_channels=1, out_channels=8, kernel_size=(3, 3), bias=False
            ),
            NeuromorphicReLU(),
            torch.nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2)),
            torch.nn.Conv2d(
                in_channels=8, out_channels=12, kernel_size=(3, 3), bias=False
            ),
            NeuromorphicReLU(),
            torch.nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2)),
            torch.nn.Conv2d(
                in_channels=12, out_channels=12, kernel_size=(3, 3), bias=False
            ),
            NeuromorphicReLU(),
            torch.nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2)),
            torch.nn.Dropout(0.5),
            torch.nn.Flatten(),
            torch.nn.Linear(432, 10, bias=False),
            NeuromorphicReLU(fanout=0),
        )


class TinyModel(torch.nn.Module):
    def __init__(self, quantize):
        super().__init__()
        self.linear = torch.nn.Linear(5, 2, bias=False)
        self.relu = NeuromorphicReLU(fanout=2, quantize=quantize)

        self.linear.weight.data = torch.tensor(
            [[1.2, 1.0, 1.0, 3.0, -2.0], [1.2, 1.0, 1.0, 2.0, -10.0]]
        )

    def forward(self, x):
        return self.relu(self.linear(x))


def test_parsing():
    model = Model()
    loss = SynOpCounter(model.modules())

    assert len(loss.modules) == 3


def test_loss():
    model = TinyModel(quantize=False)
    criterion = SynOpCounter(model.modules())
    input = torch.tensor([[0.5, 0.5, 0.5, 0.5, 0.5]])
    model(input)
    loss = criterion()

    assert np.allclose(loss.item(), 4.2)


def test_loss_quantized():
    model = TinyModel(quantize=True)
    criterion = SynOpCounter(model.modules())
    input = torch.tensor([[0.5, 0.5, 0.5, 0.5, 0.5]])
    model(input)
    loss = criterion()

    assert np.allclose(loss.item(), 4.0)


def test_layer_synops():
    model = Model()
    criterion = SynOpCounter(model.modules(), sum_activations=False)
    input = torch.rand([1, 1, 64, 64])
    model(input)
    loss = criterion()

    assert len(loss) == 3


def test_linear_synops_counter():
    model = nn.Linear(3, 5)
    input_ = torch.zeros((1, 3))
    input_[0, 0] = 3
    analyzer = SNNAnalyzer(model)
    model(input_)
    model_stats = analyzer.get_model_statistics()
    layer_stats = analyzer.get_layer_statistics()["parameter"][""]

    # 3 spikes * 5 channels
    assert model_stats["synops"] == 15
    assert layer_stats["synops"] == 15
    assert layer_stats["fanout_prev"] == 5


def test_linear_synops_counter_across_batches():
    model = nn.Linear(3, 5)
    input1 = torch.zeros((1, 3))
    input1[0, 0] = 3
    input2 = torch.zeros((1, 3))
    input2[0, 0] = 6
    analyzer = SNNAnalyzer(model)
    model(input1)
    model(input2)
    model_stats = analyzer.get_model_statistics()
    layer_stats = analyzer.get_layer_statistics()["parameter"][""]

    # (3+6)/2 spikes * 5 channels
    assert model_stats["synops"] == 22.5
    assert layer_stats["synops"] == 22.5
    assert layer_stats["fanout_prev"] == 5


def test_conv_synops_counter():
    model = nn.Conv2d(1, 5, kernel_size=2)
    input_ = torch.zeros((1, 1, 3, 3))
    input_[0, 0, 1, 1] = 3
    analyzer = SNNAnalyzer(model)
    model(input_)
    model_stats = analyzer.get_model_statistics()
    layer_stats = analyzer.get_layer_statistics()["parameter"][""]

    # 3 spikes, 2x2 kernel, 5 channels
    assert model_stats["synops"] == 60
    assert layer_stats["synops"] == 60
    assert layer_stats["fanout_prev"] == 20


def test_conv_synops_counter_counts_across_batches():
    model = nn.Conv2d(1, 5, kernel_size=2)
    input1 = torch.zeros((1, 1, 3, 3))
    input1[0, 0, 1, 1] = 6
    input2 = torch.zeros((1, 1, 3, 3))
    input2[0, 0, 1, 1] = 3
    analyzer = SNNAnalyzer(model)
    model(input1)
    model(input2)
    model_stats = analyzer.get_model_statistics()
    layer_stats = analyzer.get_layer_statistics()["parameter"][""]

    # (3+6)/2 spikes, 2x2 kernel, 5 channels
    assert model_stats["synops"] == 90
    assert layer_stats["synops"] == 90
    assert layer_stats["fanout_prev"] == 20


def test_spiking_layer_firing_rate():
    layer = sl.IAF()
    input_ = torch.eye(4).unsqueeze(0).unsqueeze(0)

    analyzer = sinabs.SNNAnalyzer(layer)
    output = layer(input_)
    model_stats = analyzer.get_model_statistics()
    layer_stats = analyzer.get_layer_statistics()["spiking"][""]

    assert (output == input_).all()
    assert model_stats["firing_rate"] == 0.25
    assert layer_stats["firing_rate"] == 0.25
    assert layer_stats["firing_rate_per_neuron"].shape == (4, 4)
    assert layer_stats["firing_rate_per_neuron"].mean() == 0.25


def test_spiking_layer_firing_rate_across_batches():
    layer = sl.IAF()
    input1 = torch.eye(4).unsqueeze(0).unsqueeze(0)
    input2 = 2 * torch.eye(4).unsqueeze(0).unsqueeze(0)

    analyzer = sinabs.SNNAnalyzer(layer)
    output = layer(input1)
    sinabs.reset_states(layer)
    output = layer(input2)
    model_stats = analyzer.get_model_statistics()
    layer_stats = analyzer.get_layer_statistics()["spiking"][""]

    assert (output == input2).all()
    assert model_stats["firing_rate"] == 0.375
    assert layer_stats["firing_rate"] == 0.375
    assert layer_stats["firing_rate_per_neuron"].shape == (4, 4)
    assert layer_stats["firing_rate_per_neuron"].mean() == 0.375


def test_analyzer_reset():
    layer = sl.IAF()
    input_ = 2 * torch.eye(4).unsqueeze(0).unsqueeze(0)

    analyzer = sinabs.SNNAnalyzer(layer)
    output = layer(input_)
    output = layer(input_)
    sinabs.reset_states(layer)
    analyzer.reset()
    output = layer(input_)
    model_stats = analyzer.get_model_statistics()
    layer_stats = analyzer.get_layer_statistics()["spiking"][""]

    assert (output == input_).all()
    assert model_stats["firing_rate"] == 0.5
    assert layer_stats["firing_rate"] == 0.5
    assert layer_stats["firing_rate_per_neuron"].shape == (4, 4)
    assert layer_stats["firing_rate_per_neuron"].mean() == 0.5


def test_snn_analyzer_statistics():
    batch_size = 3
    num_timesteps = 10
    model = nn.Sequential(
        nn.Conv2d(1, 2, kernel_size=2, bias=False),
        IAFSqueeze(batch_size=batch_size),
        nn.Conv2d(2, 3, kernel_size=2, bias=False),
        IAFSqueeze(batch_size=batch_size),
    )

    analyzer = SNNAnalyzer(model)
    input_ = torch.rand((batch_size, num_timesteps, 1, 4, 4)) * 100
    input_flattended = input_.flatten(0, 1)
    output = model(input_flattended)
    spike_layer_stats = analyzer.get_layer_statistics()["spiking"]
    param_layer_stats = analyzer.get_layer_statistics()["parameter"]
    model_stats = analyzer.get_model_statistics()

    # spiking layer checks
    assert (
        spike_layer_stats["3"]["firing_rate"] == output.mean()
    ), "The output mean should be equivalent to the firing rate of the last spiking layer"
    assert (
        torch.cat(
            (
                spike_layer_stats["1"]["firing_rate_per_neuron"].ravel(),
                spike_layer_stats["3"]["firing_rate_per_neuron"].ravel(),
            )
        ).mean()
        == model_stats["firing_rate"]
    ), "Mean of layer 1 and 3 firing rates is not equal to calculated model firing rate."

    # parameter layer checks
    param_layer_stats["0"]["synops"] == input_.mean(0).sum() * np.product(
        model[0].kernel_size
    ) * model[0].out_channels
    assert param_layer_stats["0"]["num_timesteps"] == num_timesteps
    assert param_layer_stats["2"]["num_timesteps"] == num_timesteps
    assert (
        model_stats["synops"]
        == param_layer_stats["0"]["synops"] + param_layer_stats["2"]["synops"]
    )


def test_snn_analyzer_does_not_depend_on_batch_size():
    batch_size_1 = 5
    num_timesteps = 10
    linear1 = nn.Linear(3, 4, bias=False)
    analyzer = SNNAnalyzer(linear1)
    input_ = torch.ones((batch_size_1, num_timesteps, 3)) * 10
    linear1(input_)
    model_stats_batch_size_1 = analyzer.get_model_statistics()

    batch_size_2 = 10
    linear2 = nn.Linear(3, 4, bias=False)
    analyzer = SNNAnalyzer(linear2)
    input_ = torch.ones((batch_size_2, num_timesteps, 3)) * 10
    linear2(input_)
    model_stats_batch_size_2 = analyzer.get_model_statistics()

    assert model_stats_batch_size_1["synops"] == model_stats_batch_size_2["synops"]



from copy import deepcopy
from typing import Callable
from warnings import warn

import torch.nn as nn


def replace_module(model: nn.Module, source_class: type, mapper_fn: Callable):
    """A utility function that returns a copy of the model, where specific layers are replaced with
    another type depending on the mapper function.

    Parameters:
        model: A PyTorch model.
        source_class: the layer class to replace. Each find will be passed to mapper_fn
        mapper_fn: A callable that takes as argument the layer to replace and returns the new object.

    Returns:
        A model copy with replaced modules according to mapper_fn.
    """

    # Handle case where `model` is of type `source_class`
    if type(model) == source_class:
        return mapper_fn(model)

    new_model = deepcopy(model)
    replace_module_(new_model, source_class, mapper_fn)
    return new_model


def replace_module_(model: nn.Sequential, source_class: type, mapper_fn: Callable):
    """In-place version of replace_module that will step through modules that have children and
    apply the mapper_fn.

    Parameters:
        model: A PyTorch model.
        source_class: the layer class to replace. Each find will be passed to mapper_fn
        mapper_fn: A callable that takes as argument the layer to replace and returns the new object.

    Returns:
        Modified model.
    """

    # If `model` is of type `source_class`, it cannot be converted in-place.
    if type(model) == source_class:
        warn(
            f"Provided model is of type `{source_class}` and cannot be converted"
            " in-place if not part of another Module. Apply mapper function"
            " directly or use `replace_module` to generate new object of desired type."
        )

    for name, module in model.named_children():
        if list(module.named_children()):
            replace_module_(module, source_class, mapper_fn)

        if type(module) == source_class:
            setattr(model, name, mapper_fn(module))



from typing import Callable, Optional, Union

import torch
import torch.nn as nn

from sinabs.activation import MembraneSubtract, SingleExponential, SingleSpike

from . import functional
from .reshape import SqueezeMixin
from .stateful_layer import StatefulLayer


class ALIF(StatefulLayer):
    """Adaptive Leaky Integrate and Fire neuron layer that inherits from
    :class:`~sinabs.layers.StatefulLayer`.

    Pytorch implementation of a Long Short Term Memory SNN (LSNN) by Bellec et al., 2018:
    https://papers.neurips.cc/paper/2018/hash/c203d8a151612acf12457e4d67635a95-Abstract.html

    Neuron dynamics in discrete time:

    .. math ::
        V(t+1) = \\alpha V(t) + (1-\\alpha) \\sum w.s(t)

        B(t+1) = b0 + \\text{adapt_scale } b(t)

        b(t+1) = \\rho b(t) + (1-\\rho) s(t)

        \\text{if } V_{mem}(t) >= B(t) \\text{, then } V_{mem} \\rightarrow V_{mem} - b0, b \\rightarrow 0

    where :math:`\\alpha = e^{-1/\\tau_{mem}}`, :math:`\\rho = e^{-1/\\tau_{adapt}}`
    and :math:`w.s(t)` is the input current for a spike s and weight w.

    By default there will not be any synaptic current dynamics used. You can specify tau_syn to apply an
    exponential decay kernel to the input:

    .. math ::
        i(t+1) = \\alpha_{syn} i(t) (1-\\alpha_{syn}) + input


    Parameters:
        tau_mem: Membrane potential time constant.
        tau_adapt: Spike threshold time constant.
        tau_syn: Synaptic decay time constants. If None, no synaptic dynamics are used, which is the default.
        adapt_scale: The amount that the spike threshold is bumped up for every spike, after which it decays back to the initial threshold.
        spike_threshold: Spikes are emitted if v_mem is above that threshold. By default set to 1.0.
        spike_fn: Choose a Sinabs or custom torch.autograd.Function that takes a dict of states,
                  a spike threshold and a surrogate gradient function and returns spikes. Be aware
                  that the class itself is passed here (because torch.autograd methods are static)
                  rather than an object instance.
        reset_fn: A function that defines how the membrane potential is reset after a spike.
        surrogate_grad_fn: Choose how to define gradients for the spiking non-linearity during the
                           backward pass. This is a function of membrane potential.
        min_v_mem: Lower bound for membrane potential v_mem, clipped at every time step.
        shape: Optionally initialise the layer state with given shape. If None, will be inferred from input_size.
        train_alphas: When True, the discrete decay factor exp(-1/tau) is used for training rather than tau itself.
        norm_input: When True, normalise input current by tau. This helps when training time constants.
        record_states: When True, will record all internal states such as v_mem or i_syn in a dictionary attribute `recordings`. Default is False.

    Shape:
        - Input: :math:`(Batch, Time, Channel, Height, Width)` or :math:`(Batch, Time, Channel)`
        - Output: Same as input.

    Attributes:
        v_mem: The membrane potential resets according to reset_fn for every spike.
        i_syn: This attribute is only available if tau_syn is not None.
        b: The deviation from the original spike threshold.
        spike_threshold: The current spike threshold that gets updated with every output spike.
    """

    def __init__(
        self,
        tau_mem: Union[float, torch.Tensor],
        tau_adapt: Union[float, torch.Tensor],
        tau_syn: Optional[Union[float, torch.Tensor]] = None,
        adapt_scale: Union[float, torch.Tensor] = 1.8,
        spike_threshold: torch.Tensor = torch.tensor(1.0),
        spike_fn: Callable = SingleSpike,
        reset_fn: Callable = MembraneSubtract(),
        surrogate_grad_fn: Callable = SingleExponential(),
        min_v_mem: Optional[float] = None,
        shape: Optional[torch.Size] = None,
        train_alphas: bool = False,
        norm_input: bool = True,
        record_states: bool = False,
    ):
        super().__init__(
            state_names=["v_mem", "i_syn", "b", "spike_threshold"]
            if tau_syn is not None
            else ["v_mem", "b", "spike_threshold"]
        )
        if train_alphas:
            self.alpha_mem = nn.Parameter(torch.exp(-1 / torch.as_tensor(tau_mem)))
            self.alpha_adapt = nn.Parameter(torch.exp(-1 / torch.as_tensor(tau_adapt)))
            self.alpha_syn = (
                nn.Parameter(torch.exp(-1 / torch.as_tensor(tau_syn)))
                if tau_syn is not None
                else None
            )
        else:
            self.tau_mem = nn.Parameter(torch.as_tensor(tau_mem))
            self.tau_adapt = nn.Parameter(torch.as_tensor(tau_adapt))
            self.tau_syn = nn.Parameter(torch.as_tensor(tau_syn)) if tau_syn else None
        self.adapt_scale = adapt_scale
        self.b0 = spike_threshold
        self.spike_fn = spike_fn
        self.reset_fn = reset_fn
        self.surrogate_grad_fn = surrogate_grad_fn
        self.min_v_mem = min_v_mem
        self.train_alphas = train_alphas
        self.norm_input = norm_input
        self.record_states = record_states
        if shape:
            self.init_state_with_shape(shape)

    @property
    def alpha_mem_calculated(self):
        """Calculates alpha_mem from tau_mem, if not already known."""
        return self.alpha_mem if self.train_alphas else torch.exp(-1 / self.tau_mem)

    @property
    def alpha_adapt_calculated(self):
        """Calculates alpha_adapt from tau_adapt, if not already known."""
        return self.alpha_adapt if self.train_alphas else torch.exp(-1 / self.tau_adapt)

    @property
    def alpha_syn_calculated(self):
        """Calculates alpha_syn from tau_syn, if not already known."""
        if self.train_alphas:
            return self.alpha_syn
        elif not self.train_alphas and self.tau_syn is not None:
            return torch.exp(-1 / self.tau_syn)
        else:
            return None

    def forward(self, input_data: torch.Tensor):
        """
        Parameters:
            input_data: Data to be processed. Expected shape: (batch, time, ...)

        Returns:
            Output data with same shape as `input_data`.
        """
        batch_size, time_steps, *trailing_dim = input_data.shape

        # Ensure the neuron state are initialized
        if not self.is_state_initialised() or not self.state_has_shape(
            (batch_size, *trailing_dim)
        ):
            self.init_state_with_shape((batch_size, *trailing_dim))

        alpha_mem = self.alpha_mem_calculated
        alpha_syn = self.alpha_syn_calculated
        alpha_adapt = self.alpha_adapt_calculated

        spikes, state, recordings = functional.alif_forward(
            input_data=input_data,
            alpha_mem=alpha_mem,
            alpha_adapt=alpha_adapt,
            alpha_syn=alpha_syn,
            adapt_scale=self.adapt_scale,
            state=dict(self.named_buffers()),
            spike_fn=self.spike_fn,
            reset_fn=self.reset_fn,
            surrogate_grad_fn=self.surrogate_grad_fn,
            min_v_mem=self.min_v_mem,
            b0=self.b0,
            norm_input=self.norm_input,
            record_states=self.record_states,
        )
        self.b = state["b"]
        self.v_mem = state["v_mem"]
        self.spike_threshold = state["spike_threshold"]
        self.recordings = recordings

        self.firing_rate = spikes.sum() / spikes.numel()
        return spikes

    @property
    def shape(self):
        if self.is_state_initialised():
            return self.v_mem.shape
        else:
            return None

    @property
    def _param_dict(self) -> dict:
        param_dict = super()._param_dict
        param_dict.update(
            tau_mem=-1 / torch.log(self.alpha_mem.detach_())
            if self.train_alphas
            else self.tau_mem,
            tau_adapt=-1 / torch.log(self.alpha_adapt.detach_())
            if self.train_alphas
            else self.tau_adapt,
            tau_syn=-1 / torch.log(self.alpha_syn.detach_())
            if self.train_alphas
            else self.tau_syn,
            adapt_scale=self.adapt_scale,
            spike_threshold=self.b0,
            spike_fn=self.spike_fn,
            reset_fn=self.reset_fn,
            surrogate_grad_fn=self.surrogate_grad_fn,
            train_alphas=self.train_alphas,
            shape=self.shape,
            min_v_mem=self.min_v_mem,
            record_states=self.record_states,
        )
        return param_dict


class ALIFRecurrent(ALIF):
    """Adaptive Leaky Integrate and Fire neuron layer with recurrent connections which inherits
    from :class:`~sinabs.layers.ALIF`.

    Pytorch implementation of a Long Short Term Memory SNN (LSNN) by Bellec et al., 2018:
    https://papers.neurips.cc/paper/2018/hash/c203d8a151612acf12457e4d67635a95-Abstract.html

    Neuron dynamics in discrete time:

    .. math ::
        V(t+1) = \\alpha V(t) + (1-\\alpha) \\sum w.s(t)

        B(t+1) = b0 + \\text{adapt_scale } b(t)

        b(t+1) = \\rho b(t) + (1-\\rho) s(t)

        \\text{if } V_{mem}(t) >= B(t) \\text{, then } V_{mem} \\rightarrow V_{mem} - b0, b \\rightarrow 0

    where :math:`\\alpha = e^{-1/\\tau_{mem}}`, :math:`\\rho = e^{-1/\\tau_{adapt}}`
    and :math:`w.s(t)` is the input current for a spike s and weight w.

    By default there will not be any synaptic current dynamics used. You can specify tau_syn to apply an
    exponential decay kernel to the input:

    .. math ::
        i(t+1) = \\alpha_{syn} i(t) (1-\\alpha_{syn}) + input

    Parameters:
        tau_mem: Membrane potential time constant.
        tau_adapt: Spike threshold time constant.
        rec_connect: An nn.Module which defines the recurrent connectivity, e.g. nn.Linear
        tau_syn: Synaptic decay time constants. If None, no synaptic dynamics are used, which is the default.
        adapt_scale: The amount that the spike threshold is bumped up for every spike, after which it decays back to the initial threshold.
        spike_threshold: Spikes are emitted if v_mem is above that threshold. By default set to 1.0.
        spike_fn: Choose a Sinabs or custom torch.autograd.Function that takes a dict of states,
                  a spike threshold and a surrogate gradient function and returns spikes. Be aware
                  that the class itself is passed here (because torch.autograd methods are static)
                  rather than an object instance.
        reset_fn: A function that defines how the membrane potential is reset after a spike.
        surrogate_grad_fn: Choose how to define gradients for the spiking non-linearity during the
                           backward pass. This is a function of membrane potential.
        min_v_mem: Lower bound for membrane potential v_mem, clipped at every time step.
        shape: Optionally initialise the layer state with given shape. If None, will be inferred from input_size.
        train_alphas: When True, the discrete decay factor exp(-1/tau) is used for training rather than tau itself.
        norm_input: When True, normalise input current by tau. This helps when training time constants.
        record_states: When True, will record all internal states such as v_mem or i_syn in a dictionary attribute `recordings`. Default is False.

    Shape:
        - Input: :math:`(Batch, Time, Channel, Height, Width)` or :math:`(Batch, Time, Channel)`
        - Output: Same as input.

    Attributes:
        v_mem: The membrane potential resets according to reset_fn for every spike.
        i_syn: This attribute is only available if tau_syn is not None.
        b: The deviation from the original spike threshold.
        spike_threshold: The current spike threshold that gets updated with every output spike.
    """

    def __init__(
        self,
        tau_mem: Union[float, torch.Tensor],
        tau_adapt: Union[float, torch.Tensor],
        rec_connect: torch.nn.Module,
        tau_syn: Optional[Union[float, torch.Tensor]] = None,
        adapt_scale: Union[float, torch.Tensor] = 1.8,
        spike_threshold: torch.Tensor = torch.tensor(1.0),
        spike_fn: Callable = SingleSpike,
        reset_fn: Callable = MembraneSubtract(),
        surrogate_grad_fn: Callable = SingleExponential(),
        min_v_mem: Optional[float] = None,
        shape: Optional[torch.Size] = None,
        train_alphas: bool = False,
        norm_input: bool = True,
        record_states: bool = False,
    ):
        super().__init__(
            tau_mem=tau_mem,
            tau_adapt=tau_adapt,
            tau_syn=tau_syn,
            adapt_scale=adapt_scale,
            spike_threshold=spike_threshold,
            spike_fn=spike_fn,
            reset_fn=reset_fn,
            surrogate_grad_fn=surrogate_grad_fn,
            min_v_mem=min_v_mem,
            shape=shape,
            train_alphas=train_alphas,
            norm_input=norm_input,
            record_states=record_states,
        )
        self.rec_connect = rec_connect

    def forward(self, input_data: torch.Tensor):
        """
        Parameters:
            input_data: Data to be processed. Expected shape: (batch, time, ...)

        Returns:
            Output data with same shape as `input_data`.
        """
        batch_size, time_steps, *trailing_dim = input_data.shape

        # Ensure the neuron state are initialized
        if not self.is_state_initialised() or not self.state_has_shape(
            (batch_size, *trailing_dim)
        ):
            self.init_state_with_shape((batch_size, *trailing_dim))

        alpha_mem = self.alpha_mem_calculated
        alpha_syn = self.alpha_syn_calculated
        alpha_adapt = self.alpha_adapt_calculated

        spikes, state, recordings = functional.alif_recurrent(
            input_data=input_data,
            alpha_mem=alpha_mem,
            alpha_adapt=alpha_adapt,
            alpha_syn=alpha_syn,
            adapt_scale=self.adapt_scale,
            state=dict(self.named_buffers()),
            spike_fn=self.spike_fn,
            reset_fn=self.reset_fn,
            surrogate_grad_fn=self.surrogate_grad_fn,
            min_v_mem=self.min_v_mem,
            rec_connect=self.rec_connect,
            b0=self.b0,
            norm_input=self.norm_input,
            record_states=self.record_states,
        )
        self.b = state["b"]
        self.v_mem = state["v_mem"]
        self.spike_threshold = state["spike_threshold"]
        self.recordings = recordings

        self.firing_rate = spikes.sum() / spikes.numel()
        return spikes


class ALIFSqueeze(ALIF, SqueezeMixin):
    """ALIF layer with 4-dimensional input (Batch*Time, Channel, Height, Width).

    Same as parent ALIF class, only takes in squeezed 4D input (Batch*Time, Channel, Height, Width)
    instead of 5D input (Batch, Time, Channel, Height, Width) in order to be compatible with
    layers that can only take a 4D input, such as convolutional and pooling layers.

    Shape:
        - Input: :math:`(Batch \\times Time, Channel, Height, Width)` or :math:`(Batch \\times Time, Channel)`
        - Output: Same as input.

    Attributes:
        v_mem: The membrane potential resets according to reset_fn for every spike.
        i_syn: This attribute is only available if tau_syn is not None.
        b: The deviation from the original spike threshold.
        spike_threshold: The current spike threshold that gets updated with every output spike.
    """

    def __init__(
        self,
        batch_size=None,
        num_timesteps=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.squeeze_init(batch_size, num_timesteps)

    def forward(self, input_data: torch.Tensor) -> torch.Tensor:
        """Forward call wrapper that will flatten the input to and unflatten the output from the
        super class forward call."""
        return self.squeeze_forward(input_data, super().forward)

    @property
    def _param_dict(self) -> dict:
        return self.squeeze_param_dict(super()._param_dict)



import torch
import torch.nn as nn

import sinabs
import sinabs.layers as sl


class SNN(nn.Module):
    def __init__(self, hidden_dim=10):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(2, hidden_dim),
            sl.IAF(),
            nn.Linear(hidden_dim, hidden_dim),
            sl.IAF(),
            nn.Linear(hidden_dim, hidden_dim),
            sl.IAF(),
            nn.Linear(hidden_dim, 5),
        )
        self.spike_output = sl.IAF()

    def forward(self, x):
        return self.spike_output(self.net(x))


def test_reset_states():
    model = SNN()
    assert model.net[1].v_mem.sum() == 0

    input = torch.rand((5, 10, 2))
    model(input)
    assert model.net[1].v_mem.grad_fn is not None
    assert model.net[1].v_mem.sum() != 0

    sinabs.reset_states(model)
    assert model.net[1].v_mem.grad_fn is None
    assert model.net[1].v_mem.sum() == 0


def test_zero_grad():
    model = SNN()
    assert model.net[1].v_mem.grad == None

    input = torch.rand((5, 10, 2))
    output = model(input)

    assert model.net[1].v_mem.grad_fn is not None
    assert model.net[1].v_mem.sum() != 0

    sinabs.zero_grad(model)
    assert model.net[1].v_mem.grad_fn is None
    assert model.net[1].v_mem.sum() != 0



from typing import Dict, List, Optional, Tuple

import torch


class StatefulLayer(torch.nn.Module):
    """A base class that instantiates buffers/states which update at every time step and provides
    helper methods that manage those states.

    Parameters:
        state_names: the PyTorch buffers to initialise. These are not parameters.
    """

    def __init__(self, state_names: List[str]):
        super().__init__()

        for state_name in state_names:
            self.register_buffer(state_name, torch.zeros((0)))

    def zero_grad(self, set_to_none: bool = False) -> None:
        r"""
        Zero's the gradients for buffers/state along with the parameters.
        See :meth:`torch.nn.Module.zero_grad` for details
        """
        # Zero grad parameters
        super().zero_grad(set_to_none)
        if self.is_state_initialised():
            # Zero grad buffers
            for b in self.buffers():
                if b.grad_fn is not None:
                    b.detach_()
                else:
                    b.requires_grad_(False)

    def forward(self, *args, **kwargs):
        """
        Not implemented - You need to implement a forward method in child class
        """
        raise NotImplementedError(
            "No forward method has been implemented for this class"
        )

    def is_state_initialised(self) -> bool:
        """Checks if buffers are of shape 0 and returns True only if none of them are."""
        for buffer in self.buffers():
            if buffer.shape == torch.Size([0]):
                return False
        return True

    def state_has_shape(self, shape) -> bool:
        """Checks if all state have a given shape."""
        for buff in self.buffers():
            if buff.shape != shape:
                return False
        return True

    def init_state_with_shape(self, shape, randomize: bool = False) -> None:
        """Initialise state/buffers with either zeros or random tensor of specific shape."""
        for name, buffer in self.named_buffers():
            self.register_buffer(name, torch.zeros(shape, device=buffer.device))
        self.reset_states(randomize=randomize)

    def reset_states(
        self,
        randomize: bool = False,
        value_ranges: Optional[Dict[str, Tuple[float, float]]] = None,
    ):
        """Reset the state/buffers in a layer.

        Parameters:
            randomize: If true, reset the states between a range provided. Else, the states are reset to zero.
            value_ranges: A dictionary of key value pairs: buffer_name -> (min, max) for each state that needs to be reset.
                          The states are reset with a uniform distribution between the min and max values specified.
                          Any state with an undefined key in this dictionary will be reset between 0 and 1
                          This parameter is only used if randomize is set to true.

        .. note:: If you would like to reset the state with a custom distribution, you can do this individually for each parameter as follows::

            layer.<state_name>.data = <your desired data>

            layer.<state_name>.detach_()
        """
        if self.is_state_initialised():
            for name, buffer in self.named_buffers():
                if randomize:
                    if value_ranges and name in value_ranges:
                        min_value, max_value = value_ranges[name]
                    else:
                        min_value, max_value = (0.0, 1.0)
                    # Initialize with uniform distribution
                    torch.nn.init.uniform_(buffer)
                    # Rescale the value
                    buffer.data = buffer * (max_value - min_value) + min_value
                else:
                    buffer.zero_()
                buffer.detach_()

    def __repr__(self):
        param_strings = [
            f"{key}={value}"
            for key, value in self._param_dict.items()
            if key
            in [
                "tau_mem",
                "tau_syn",
                "tau_adapt",
                "adapt_scale",
                "spike_threshold",
                "min_v_mem",
                "norm_input",
                "batch_size",
                "num_timesteps",
            ]
            and value is not None
        ]
        param_strings = ", ".join(param_strings)
        return f"{self.__class__.__name__}({param_strings})"

    def __deepcopy__(self, memo=None):
        copy = self.__class__(**self._param_dict)

        # Copy parameters
        for name, param in self.named_parameters():
            new_inst_param = getattr(copy, name)
            new_inst_param.data = param.data.clone()
        # Copy buffers (using state dict will fail if buffers have non-default shapes)
        for name, buffer in self.named_buffers():
            new_inst_buffer = getattr(copy, name)
            new_inst_buffer.data = buffer.data.clone()  # Copy parameters

        return copy

    @property
    def _param_dict(self) -> dict:
        """Dict of all parameters relevant for creating a new instance with same parameters as
        `self`."""
        return dict()

    @property
    def arg_dict(self) -> dict:
        """A public getter function for the constructor arguments."""
        return self._param_dict

    @property
    def does_spike(self) -> bool:
        """Return True if the layer has an activation function."""
        return hasattr(self, "spike_fn") and self.spike_fn is not None



import math
from dataclasses import dataclass

import torch


@dataclass
class Heaviside:
    """Heaviside surrogate gradient with optional shift.

    Parameters:
        window: Distance between step of Heaviside surrogate gradient and
                threshold, relative to threshold.
    """

    window: float = 1.0

    def __call__(self, v_mem, spike_threshold):
        return ((v_mem >= (spike_threshold - self.window)).float()) / spike_threshold


def gaussian(x: torch.Tensor, mu: float, sigma: float):
    return torch.exp(-((x - mu) ** 2) / (2 * sigma**2)) / (
        sigma * torch.sqrt(2 * torch.tensor(math.pi))
    )


@dataclass
class Gaussian:
    """Gaussian surrogate gradient function.

    Parameters
        mu: The mean of the Gaussian.
        sigma: The standard deviation of the Gaussian.
        grad_scale: Scale the gradients arbitrarily.
    """

    mu: float = 0.0
    sigma: float = 0.5
    grad_scale: float = 1.0

    def __call__(self, v_mem, spike_threshold):
        return (
            gaussian(x=v_mem - spike_threshold, mu=self.mu, sigma=self.sigma)
            * self.grad_scale
        )


@dataclass
class MultiGaussian:
    """Surrogate gradient as defined in Yin et al., 2021.

    https://www.biorxiv.org/content/10.1101/2021.03.22.436372v2

    Parameters
        mu: The mean of the Gaussian.
        sigma: The standard deviation of the Gaussian.
        h: Controls the magnitude of the negative parts of the kernel.
        s: Controls the width of the negative parts of the kernel.
        grad_scale: Scale the gradients arbitrarily.
    """

    mu: float = 0.0
    sigma: float = 0.5
    h: float = 0.15
    s: float = 6
    grad_scale: float = 1.0

    def __call__(self, v_mem, spike_threshold):
        return (
            (1 + self.h)
            * gaussian(x=v_mem - spike_threshold, mu=self.mu, sigma=self.sigma)
            - self.h
            * gaussian(
                x=v_mem - spike_threshold, mu=self.sigma, sigma=self.s * self.sigma
            )
            - self.h
            * gaussian(
                x=v_mem - spike_threshold, mu=-self.sigma, sigma=self.s * self.sigma
            )
        ) * self.grad_scale


@dataclass
class SingleExponential:
    """Surrogate gradient as defined in Shrestha and Orchard, 2018.

    https://papers.nips.cc/paper/2018/hash/82f2b308c3b01637c607ce05f52a2fed-Abstract.html
    """

    grad_width: float = 0.5
    grad_scale: float = 1.0

    def __call__(self, v_mem, spike_threshold):
        abs_width = spike_threshold * self.grad_width
        return (
            self.grad_scale
            / abs_width
            * torch.exp(-torch.abs(v_mem - spike_threshold) / abs_width)
        )


@dataclass
class PeriodicExponential:
    """Surrogate gradient as defined in Weidel and Sheik, 2021.

    https://arxiv.org/abs/2111.01456
    """

    grad_width: float = 0.5
    grad_scale: float = 1.0

    def __call__(self, v_mem, spike_threshold):
        vmem_shifted = v_mem - spike_threshold / 2
        vmem_periodic = vmem_shifted - torch.div(
            vmem_shifted, spike_threshold, rounding_mode="floor"
        )
        vmem_below = vmem_shifted * (v_mem < spike_threshold)
        vmem_above = vmem_periodic * (v_mem >= spike_threshold)
        vmem_new = vmem_above + vmem_below
        spikePdf = (
            torch.exp(-torch.abs(vmem_new - spike_threshold / 2) / self.grad_width)
            / spike_threshold
        )

        return self.grad_scale * spikePdf



"""
=============
MultiGaussian
=============
:class:`~sinabs.activation.MultiGaussian` surrogate gradient.
"""

import matplotlib.pyplot as plt
import torch

import sinabs.activation as sina

x = torch.linspace(-2, 4, 500)
plt.plot(x, sina.MultiGaussian()(v_mem=x, spike_threshold=1), label="MultiGaussian")
plt.xlabel("Neuron membrane potential")
plt.ylabel("Derivative")
plt.legend()
plt.show()



from typing import List, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn

from sinabs.cnnutils import conv_output_size

# - Type alias for array-like objects
ArrayLike = Union[np.ndarray, List, Tuple]


class SpikingMaxPooling2dLayer(nn.Module):
    """Torch implementation of SpikingMaxPooling."""

    def __init__(
        self,
        pool_size: ArrayLike,
        strides: Optional[ArrayLike] = None,
        padding: ArrayLike = (0, 0, 0, 0),
        # state_number: int = 16,
    ):
        super().__init__()
        self.padding = padding
        self.pool_size = pool_size
        if strides is None:
            strides = pool_size
        self.strides = strides
        if padding == (0, 0, 0, 0):
            self.pad = None
        else:
            self.pad = nn.ZeroPad2d(padding)
        self.pool = nn.MaxPool2d(kernel_size=pool_size, stride=strides)
        # self.state_number = state_number

        # Blank parameter place holders
        self.spikes_number = None

    def forward(self, binary_input):
        # Determine no. of time steps from input
        time_steps = len(binary_input)

        # Calculate the cumulative sum spikes of each neuron
        sum_count = torch.cumsum(binary_input, 0)

        # max_sum is the pooled sum_count
        if self.pad is None:
            max_sum = self.pool(sum_count)
        else:
            max_sum = self.pool(self.pad(sum_count))

        # make sure a single spike, how much sum_count it brings, max_input_sum shows that
        input_sum = sum_count * (binary_input > 0).float()
        # pool all inputs at once
        if self.pad is None:
            max_input_sum = self.pool(input_sum)
        else:
            max_input_sum = self.pool(self.pad(input_sum))

        # max of 1 and 0 s from spike train
        if self.pad is None:
            original_max_input_sum = self.pool(binary_input)
        else:
            original_max_input_sum = self.pool(self.pad(binary_input))

        # Make sure the max sum is brought by the single spike from input_sum
        # (max_input_sum >= max_sum).float() is the gate to pass through spikes
        max_input_sum = (max_input_sum >= max_sum).float() * original_max_input_sum

        self.spikes_number = max_input_sum.abs().sum()
        self.tw = len(max_input_sum)
        return max_input_sum.float()  # Float is just to keep things compatible

    def get_output_shape(self, input_shape: Tuple) -> Tuple:
        """Returns the shape of output, given an input to this layer.

        Parameters:
            input_shape: (channels, height, width)

        Returns:
            (channelsOut, height_out, width_out)
        """
        (channels, height, width) = input_shape

        height_out = conv_output_size(
            height + sum(self.padding[2:]), self.pool_size[0], self.strides[0]
        )
        width_out = conv_output_size(
            width + sum(self.padding[:2]), self.pool_size[1], self.strides[1]
        )
        return channels, height_out, width_out


class SumPool2d(torch.nn.LPPool2d):
    """Non-spiking sumpooling layer to be used in analogue Torch models. It is identical to
    torch.nn.LPPool2d with p=1.

    Parameters:
        kernel_size: the size of the window
        stride: the stride of the window. Default value is kernel_size
        ceil_mode: when True, will use ceil instead of floor to compute the output shape
    """

    def __init__(self, kernel_size, stride=None, ceil_mode=False):
        super().__init__(1, kernel_size, stride, ceil_mode)



import torch

from sinabs.layers import NeuromorphicReLU


def test_forward():
    t_inp = torch.tensor([-5.5, -3.23, 2.3, 0.0])
    relu = NeuromorphicReLU()
    t_out = relu(t_inp)

    assert (t_out == torch.tensor([0.0, 0.0, 2.0, 0.0])).all()


def test_backward():
    t_inp = torch.tensor([-5.5, -3.23, 2.3, 0.1], requires_grad=True)
    relu = NeuromorphicReLU()
    t_out = relu(t_inp)
    z = 2 * t_out.sum()
    z.backward()
    grad = t_inp.grad

    assert (grad == torch.tensor([0.0, 0.0, 2.0, 2.0])).all()



import torch

from sinabs.layers import ExpLeak, ExpLeakSqueeze


def test_leaky_basic():
    time_steps = 100
    tau_mem = torch.tensor(30.0)
    input_current = torch.rand(time_steps, 2, 7, 7)
    layer = ExpLeak(tau_mem=tau_mem)
    membrane_output = layer(input_current)

    assert not layer.does_spike
    assert input_current.shape == membrane_output.shape
    assert torch.isnan(membrane_output).sum() == 0
    assert membrane_output.sum() > 0


def test_leaky_squeezed():
    batch_size = 10
    time_steps = 100
    tau_mem = torch.tensor(30.0)
    input_current = torch.rand(batch_size * time_steps, 2, 7, 7)
    layer = ExpLeakSqueeze(tau_mem=tau_mem, batch_size=batch_size)
    membrane_output = layer(input_current)

    assert input_current.shape == membrane_output.shape
    assert torch.isnan(membrane_output).sum() == 0
    assert membrane_output.sum() > 0


def test_leaky_membrane_decay():
    batch_size = 10
    time_steps = 100
    tau_mem = torch.tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.zeros(batch_size, time_steps, 2, 7, 7)
    input_current[:, 0] = 1 / (1 - alpha)  # only inject current in the first time step
    layer = ExpLeak(tau_mem=tau_mem, norm_input=True)
    membrane_output = layer(input_current)

    # first time step is not decayed
    membrane_decay = alpha ** (time_steps - 1)

    # account for rounding errors with .isclose()
    assert torch.allclose(
        membrane_output[:, 0], torch.tensor(1.0)
    ), "Output for first time step is not correct."
    assert (
        membrane_output[:, -1] == layer.v_mem
    ).all(), "Output of last time step does not correspond to last layer state."
    assert torch.isclose(
        layer.v_mem, membrane_decay, atol=1e-08
    ).all(), "Neuron membrane potentials do not seems to decay correctly."


def test_leaky_v_mem_recordings():
    batch_size, time_steps = 10, 100
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7)
    layer = ExpLeak(tau_mem=30.0, record_states=True)
    membrane_output = layer(input_current)

    layer.recordings["v_mem"].sum().backward()
    assert layer.tau_mem.grad is not None

    assert layer.recordings["v_mem"].shape == membrane_output.shape
    assert "i_syn" not in layer.recordings.keys()



from .alif import ALIF, ALIFRecurrent
from .crop2d import Cropping2dLayer
from .exp_leak import ExpLeak, ExpLeakSqueeze
from .iaf import IAF, IAFRecurrent, IAFSqueeze
from .lif import LIF, LIFRecurrent, LIFSqueeze
from .neuromorphic_relu import NeuromorphicReLU
from .pool2d import SpikingMaxPooling2dLayer, SumPool2d
from .quantize import QuantizeLayer
from .reshape import FlattenTime, Repeat, SqueezeMixin, UnflattenTime
from .stateful_layer import StatefulLayer
from .to_spike import Img2SpikeLayer, Sig2SpikeLayer



from typing import Callable, Optional

import numpy as np
import torch

from sinabs.activation import MembraneSubtract, MultiSpike, SingleExponential

from .lif import LIF, LIFRecurrent
from .reshape import SqueezeMixin


class IAF(LIF):
    """Integrate and Fire neuron layer that is designed as a special case of
    :class:`~sinabs.layers.LIF` with tau_mem=inf.

    Neuron dynamics in discrete time:

    .. math ::
        V_{mem}(t+1) = V_{mem}(t) + \\sum z(t)

        \\text{if } V_{mem}(t) >= V_{th} \\text{, then } V_{mem} \\rightarrow V_{reset}

    where :math:`\\sum z(t)` represents the sum of all input currents at time :math:`t`.

    Parameters:
        spike_threshold: Spikes are emitted if v_mem is above that threshold. By default set to 1.0.
        spike_fn: Choose a Sinabs or custom torch.autograd.Function that takes a dict of states,
                  a spike threshold and a surrogate gradient function and returns spikes. Be aware
                  that the class itself is passed here (because torch.autograd methods are static)
                  rather than an object instance.
        reset_fn: A function that defines how the membrane potential is reset after a spike.
        surrogate_grad_fn: Choose how to define gradients for the spiking non-linearity during the
                           backward pass. This is a function of membrane potential.
        tau_syn: Synaptic decay time constants. If None, no synaptic dynamics are used, which is the default.
        min_v_mem: Lower bound for membrane potential v_mem, clipped at every time step.
        shape: Optionally initialise the layer state with given shape. If None, will be inferred from input_size.
        record_states: When True, will record all internal states such as v_mem or i_syn in a dictionary attribute
                       `recordings`. Default is False.

    Shape:
        - Input: :math:`(Batch, Time, Channel, Height, Width)` or :math:`(Batch, Time, Channel)`
        - Output: Same as input.

    Attributes:
        v_mem: The membrane potential resets according to reset_fn for every spike.
        i_syn: This attribute is only available if tau_syn is not None.
    """

    def __init__(
        self,
        spike_threshold: torch.Tensor = torch.tensor(1.0),
        spike_fn: Callable = MultiSpike,
        reset_fn: Callable = MembraneSubtract(),
        surrogate_grad_fn: Callable = SingleExponential(),
        tau_syn: Optional[float] = None,
        min_v_mem: Optional[float] = None,
        shape: Optional[torch.Size] = None,
        record_states: bool = False,
    ):
        super().__init__(
            tau_mem=np.inf,
            tau_syn=tau_syn,
            spike_threshold=spike_threshold,
            spike_fn=spike_fn,
            reset_fn=reset_fn,
            surrogate_grad_fn=surrogate_grad_fn,
            min_v_mem=min_v_mem,
            shape=shape,
            norm_input=False,
            record_states=record_states,
        )
        # IAF does not have time constants
        self.tau_mem = None

    @property
    def alpha_mem_calculated(self):
        """Always returns a tensor of 1."""
        return torch.tensor(1.0)

    @property
    def _param_dict(self) -> dict:
        param_dict = super()._param_dict
        param_dict.pop("tau_mem")
        param_dict.pop("train_alphas")
        param_dict.pop("norm_input")
        return param_dict


class IAFRecurrent(LIFRecurrent):
    """Integrate and Fire neuron layer with recurrent connections which inherits from
    :class:`~sinabs.layers.LIFRecurrent`.

    Neuron dynamics in discrete time:

    .. math ::
        V_{mem}(t+1) = V_{mem}(t) + \\sum z(t)

        \\text{if } V_{mem}(t) >= V_{th} \\text{, then } V_{mem} \\rightarrow V_{reset}

    where :math:`\\sum z(t)` represents the sum of all input currents at time :math:`t`.

    Parameters:
        rec_connect: An nn.Module which defines the recurrent connectivity, e.g. nn.Linear
        spike_threshold: Spikes are emitted if v_mem is above that threshold. By default set to 1.0.
        spike_fn: Choose a Sinabs or custom torch.autograd.Function that takes a dict of states,
                  a spike threshold and a surrogate gradient function and returns spikes. Be aware
                  that the class itself is passed here (because torch.autograd methods are static)
                  rather than an object instance.
        reset_fn: A function that defines how the membrane potential is reset after a spike.
        surrogate_grad_fn: Choose how to define gradients for the spiking non-linearity during the
                           backward pass. This is a function of membrane potential.
        tau_syn: Synaptic decay time constants. If None, no synaptic dynamics are used, which is the default.
        min_v_mem: Lower bound for membrane potential v_mem, clipped at every time step.
        shape: Optionally initialise the layer state with given shape. If None, will be inferred from input_size.
        record_states: When True, will record all internal states such as v_mem or i_syn in a dictionary attribute `recordings`. Default is False.

    Shape:
        - Input: :math:`(Batch, Time, Channel, Height, Width)` or :math:`(Batch, Time, Channel)`
        - Output: Same as input.

    Attributes:
        v_mem: The membrane potential resets according to reset_fn for every spike.
        i_syn: This attribute is only available if tau_syn is not None.
    """

    def __init__(
        self,
        rec_connect: torch.nn.Module,
        spike_threshold: torch.Tensor = torch.tensor(1.0),
        spike_fn: Callable = MultiSpike,
        reset_fn: Callable = MembraneSubtract(),
        surrogate_grad_fn: Callable = SingleExponential(),
        tau_syn: Optional[float] = None,
        min_v_mem: Optional[float] = None,
        shape: Optional[torch.Size] = None,
        record_states: bool = False,
    ):
        super().__init__(
            rec_connect=rec_connect,
            tau_mem=np.inf,
            tau_syn=tau_syn,
            spike_threshold=spike_threshold,
            spike_fn=spike_fn,
            reset_fn=reset_fn,
            surrogate_grad_fn=surrogate_grad_fn,
            min_v_mem=min_v_mem,
            shape=shape,
            norm_input=False,
            record_states=record_states,
        )
        # IAF does not have time constants
        self.tau_mem = None

    @property
    def alpha_mem_calculated(self):
        """Always returns a tensor of 1."""
        return torch.tensor(1.0)

    @property
    def _param_dict(self) -> dict:
        param_dict = super()._param_dict
        param_dict.pop("tau_mem")
        param_dict.pop("train_alphas")
        param_dict.pop("norm_input")
        return param_dict


class IAFSqueeze(IAF, SqueezeMixin):
    """IAF layer with 4-dimensional input (Batch*Time, Channel, Height, Width).

    Same as parent IAF class, only takes in squeezed 4D input (Batch*Time, Channel, Height, Width)
    instead of 5D input (Batch, Time, Channel, Height, Width) in order to be compatible with
    layers that can only take a 4D input, such as convolutional and pooling layers.

    Shape:
        - Input: :math:`(Batch \\times Time, Channel, Height, Width)` or :math:`(Batch \\times Time, Channel)`
        - Output: Same as input.

    Attributes:
        v_mem: The membrane potential resets according to reset_fn for every spike.
        i_syn: This attribute is only available if tau_syn is not None.
    """

    def __init__(
        self,
        batch_size=None,
        num_timesteps=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.squeeze_init(batch_size, num_timesteps)

    def forward(self, input_data: torch.Tensor) -> torch.Tensor:
        """Forward call wrapper that will flatten the input to and unflatten the output from the
        super class forward call."""
        return self.squeeze_forward(input_data, super().forward)

    @property
    def _param_dict(self) -> dict:
        return self.squeeze_param_dict(super()._param_dict)



from .quantize import Quantize, StochasticRounding
from .reset_mechanism import MembraneReset, MembraneSubtract
from .spike_generation import MaxSpike, MultiSpike, SingleSpike
from .surrogate_gradient_fn import (
    Gaussian,
    Heaviside,
    MultiGaussian,
    PeriodicExponential,
    SingleExponential,
)



import torch
import torch.nn as nn
from packaging import version

import sinabs.layers as sl


def test_repeat():
    data = torch.rand((5, 100, 2, 10, 10))
    conv = nn.Conv2d(2, 4, 10)
    repeated_conv = sl.Repeat(conv)
    output = repeated_conv(data)
    alt_output = conv(data.flatten(0, 1)).unflatten(0, (5, 100))

    if version.parse(torch.__version__) > version.parse("1.12"):
        torch.testing.assert_close(output, alt_output)
    else:
        torch.testing.assert_allclose(output, alt_output)


def test_flatten_time():
    data = torch.rand((5, 100, 2, 10, 10))
    flatten = sl.FlattenTime()
    output = flatten(data)

    assert output.shape == (500, 2, 10, 10)


def test_unflatten_time():
    data = torch.rand((500, 2, 10, 10))
    unflatten = sl.UnflattenTime(batch_size=5)
    output = unflatten(data)

    assert output.shape == (5, 100, 2, 10, 10)



import numpy as np
import pytest
import torch
import torch.nn as nn

from sinabs.layers import ALIF, ALIFRecurrent


def test_alif_basic():
    batch_size, time_steps = 10, 100
    tau_mem = torch.as_tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = ALIF(tau_mem=tau_mem, tau_adapt=tau_mem, adapt_scale=1.8)
    spike_output = layer(input_current)

    assert layer.does_spike
    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_alif_basic_audio_input():
    batch_size, time_steps = 10, 100
    tau_mem = torch.as_tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 64) / (1 - alpha)
    layer = ALIF(tau_mem=tau_mem, tau_adapt=tau_mem, adapt_scale=1.8)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_alif_minimum_spike_threshold():
    batch_size, time_steps = 10, 100
    tau_mem = torch.as_tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = ALIF(
        tau_mem=tau_mem,
        tau_adapt=tau_mem,
    )
    spike_output = layer(input_current)

    assert (
        layer.spike_threshold + layer.b >= 1.0
    ).all(), "Spike thresholds should not drop below initital threshold."
    assert spike_output.sum() > 0
    assert (
        layer.spike_threshold + layer.b > 1.0
    ).all(), "Spike thresholds should be above default threshold if any spikes occured."


def test_alif_spike_threshold_decay():
    batch_size, time_steps = 10, 100
    tau_mem = torch.as_tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    spike_threshold = 1.0
    input_current = torch.zeros(batch_size, time_steps, 2, 7, 7)
    input_current[:, 0] = spike_threshold / (
        1 - alpha
    )  # only inject current in the first time step and make it spike
    layer = ALIF(tau_mem=tau_mem, tau_adapt=tau_mem, adapt_scale=1 / (1 - alpha))
    spike_output = layer(input_current)
    layer.reset_states()
    layer.b.fill_(0.0)
    spike_output = layer(input_current)

    assert (layer.spike_threshold > spike_threshold).all()
    assert (
        spike_output.sum()
        == torch.prod(torch.as_tensor(input_current.size())) / time_steps
    ), "All neurons should spike exactly once."
    # decay only starts after first time step
    threshold_decay = alpha ** (time_steps - 1)
    # account for rounding errors with .isclose()
    assert torch.isclose(
        layer.spike_threshold - spike_threshold, threshold_decay, atol=1e-08
    ).all(), "Neuron spike thresholds do not seems to decay correctly."


def test_alif_with_current_dynamics():
    batch_size, time_steps = 10, 100
    tau_mem = torch.as_tensor(30.0)
    tau_syn = torch.tensor(10.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = ALIF(tau_mem=tau_mem, tau_syn=tau_syn, tau_adapt=tau_mem)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_alif_train_alphas():
    batch_size, time_steps = 10, 100
    tau_mem = torch.as_tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = ALIF(tau_mem=tau_mem, train_alphas=True, tau_adapt=tau_mem)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_alif_train_alphas_with_current_dynamics():
    batch_size, time_steps = 10, 100
    tau_mem = torch.as_tensor(30.0)
    tau_syn = torch.tensor(10.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = ALIF(tau_mem=tau_mem, tau_syn=tau_syn, tau_adapt=tau_mem, train_alphas=True)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_alif_recurrent():
    batch_size, time_steps = 10, 100
    tau_mem = torch.as_tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_dimensions = (batch_size, time_steps, 2, 10)
    n_neurons = np.product(input_dimensions[2:])
    input_current = torch.ones(*input_dimensions) * 0.5 / (1 - alpha)

    rec_connect = nn.Sequential(
        nn.Flatten(), nn.Linear(n_neurons, n_neurons, bias=False)
    )
    rec_connect[1].weight = nn.Parameter(
        torch.ones(n_neurons, n_neurons) / n_neurons * 0.5 / (1 - alpha)
    )
    layer = ALIFRecurrent(tau_mem=tau_mem, tau_adapt=tau_mem, rec_connect=rec_connect)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_alif_v_mem_recordings():
    batch_size, time_steps = 10, 100
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7)
    layer = ALIF(tau_mem=20.0, tau_adapt=10.0, norm_input=False, record_states=True)
    spike_output = layer(input_current)

    layer.recordings["v_mem"].sum().backward()
    assert layer.tau_mem.grad is not None

    assert layer.recordings["v_mem"].shape == spike_output.shape
    assert layer.recordings["spike_threshold"].shape == spike_output.shape
    assert "i_syn" not in layer.recordings.keys()


def test_alif_i_syn_recordings():
    batch_size, time_steps = 10, 100
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7)
    layer = ALIF(
        tau_mem=20.0, tau_adapt=10.0, tau_syn=10.0, norm_input=False, record_states=True
    )
    spike_output = layer(input_current)

    layer.recordings["i_syn"].sum().backward()
    assert layer.tau_syn.grad is not None
    assert layer.tau_mem.grad is None

    assert layer.recordings["v_mem"].shape == spike_output.shape
    assert layer.recordings["i_syn"].shape == spike_output.shape


def test_alif_recurrent_v_mem_recordings():
    batch_size, time_steps, n_neurons = 10, 100, 20
    input_current = torch.rand(batch_size, time_steps, n_neurons)
    rec_connect = nn.Sequential(
        nn.Flatten(), nn.Linear(n_neurons, n_neurons, bias=False)
    )
    layer = ALIFRecurrent(
        tau_mem=20.0,
        tau_adapt=10.0,
        rec_connect=rec_connect,
        norm_input=False,
        record_states=True,
    )
    spike_output = layer(input_current)

    layer.recordings["v_mem"].sum().backward()
    assert layer.tau_mem.grad is not None

    assert layer.recordings["v_mem"].shape == spike_output.shape
    assert "i_syn" not in layer.recordings.keys()


def test_alif_recurrent_i_syn_recordings():
    batch_size, time_steps, n_neurons = 10, 100, 20
    input_current = torch.rand(batch_size, time_steps, n_neurons)
    rec_connect = nn.Sequential(
        nn.Flatten(), nn.Linear(n_neurons, n_neurons, bias=False)
    )
    layer = ALIFRecurrent(
        tau_mem=20.0,
        tau_adapt=10.0,
        tau_syn=10.0,
        rec_connect=rec_connect,
        norm_input=False,
        record_states=True,
    )
    spike_output = layer(input_current)

    layer.recordings["i_syn"].sum().backward()
    assert layer.tau_syn.grad is not None

    assert layer.recordings["v_mem"].shape == spike_output.shape
    assert layer.recordings["i_syn"].shape == spike_output.shape


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_alif_on_gpu():
    batch_size, time_steps = 10, 100
    tau_mem = torch.as_tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = ALIF(tau_mem=tau_mem, tau_adapt=tau_mem)

    layer = layer.to("cuda")
    layer.reset_states()
    layer(input_current.to("cuda"))


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_alif_recurrent_on_gpu():
    batch_size, time_steps, n_neurons = 10, 100, 5
    tau_mem = torch.as_tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.zeros((batch_size, time_steps, n_neurons))
    input_current[:, 0] = 1 / (1 - alpha)

    rec_connect = nn.Linear(n_neurons, n_neurons, bias=False)
    layer = ALIFRecurrent(tau_mem=tau_mem, tau_adapt=tau_mem, rec_connect=rec_connect)

    layer = layer.to("cuda")
    layer(input_current.to("cuda"))



import numpy as np
import torch
from torch import nn

from sinabs.from_torch import from_model


class NestedANN(nn.Module):
    def __init__(self):
        super().__init__()
        seq = [
            nn.Conv2d(
                2, 8, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0), bias=False
            ),
            nn.ReLU(),
            nn.AvgPool2d(kernel_size=(2, 2)),
            nn.Conv2d(
                8, 16, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False
            ),
            nn.ReLU(),
            nn.AvgPool2d(kernel_size=(2, 2)),
            nn.Dropout(0.5),
            nn.Conv2d(
                16, 8, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False
            ),
            nn.ReLU(),
            nn.AvgPool2d(kernel_size=(2, 2)),
            nn.Flatten(),
            nn.Dropout(0.5),
            nn.Linear(8 * 8 * 8, 11, bias=True),
            nn.ReLU(),
        ]
        self.main = nn.Sequential(*seq)

    def forward(self, x):
        return self.main(x)


# init nested_ann instance for test
nested_ann = NestedANN()
input_shape = (2, 128, 128)
nested_input_tensor = torch.rand([1, 2, 128, 128])
nested_network = from_model(nested_ann, input_shape=input_shape, batch_size=1)

# init torch.nn.Sequential instance for test
ann = nn.Sequential(nn.Conv2d(1, 16, kernel_size=(3, 3), bias=False), nn.ReLU())
network = from_model(ann, batch_size=1)
data = torch.rand((1, 1, 4, 4))


def compare_networks(net0, net1):
    for module0, module1 in zip(net0.modules(), net1.modules()):
        for p0, p1 in zip(module0.parameters(), module1.parameters()):
            assert (p0 == p1).all()
        for b0, b1 in zip(module0.buffers(), module1.buffers()):
            assert (b0 == b1).all()
        if hasattr(module0, "_param_dict"):
            neuron_params1 = module1._param_dict
            for key, val in module0._param_dict.items():
                if key in neuron_params1:
                    assert neuron_params1[key] == val


def test_compare_activations():
    analog, rates, layer_names = network.compare_activations(data)
    assert len(analog) == len(rates) == len(layer_names) == 2
    for anl, spk in zip(analog, rates):
        assert np.squeeze(anl).shape == np.squeeze(spk).shape

    # Test nested_ann
    # get the names of all spiking layers
    input_layers_list = [
        name
        for name, module in nested_ann.named_modules()
        if isinstance(module, nn.ReLU)
    ]

    # test with a given layer names list
    analog, rates, returned_layer_list = nested_network.compare_activations(
        nested_input_tensor, input_layers_list
    )
    assert len(returned_layer_list) == len(input_layers_list)
    for returned_layer, input_layer in zip(returned_layer_list, input_layers_list):
        assert returned_layer == input_layer
    for anl, spk in zip(analog, rates):
        assert np.squeeze(anl).shape == np.squeeze(spk).shape

    # test when not-given a layer names list
    analog, rates, returned_layer_names = nested_network.compare_activations(
        nested_input_tensor
    )
    input_layers_list.insert(
        0, "Input"
    )  # insert a layer named "Input" in the 1st place
    assert len(returned_layer_list) == len(input_layers_list)
    for returned_layer, input_layer in zip(returned_layer_list, input_layers_list):
        assert returned_layer == input_layer
    for anl, spk in zip(analog, rates):
        assert np.squeeze(anl).shape == np.squeeze(spk).shape


def test_plot_comparison():
    """Test whether the plot_comparison() method of the sinabs.network.Network class could plot a
    nested-network which is not defined by torch.nn.Sequential(*module_list) directly."""

    # get the names of all spiking layers
    spiking_layers_names = [
        name
        for name, module in nested_ann.named_modules()
        if isinstance(module, nn.ReLU)
    ]
    num_spiking_layer = len(spiking_layers_names)

    ann_activation, snn_activation = nested_network.plot_comparison(nested_input_tensor)
    # if name_list arg not given, it will plot all spiking layer and 1 input layer
    assert len(ann_activation) == (num_spiking_layer + 1)
    assert len(snn_activation) == (num_spiking_layer + 1)
    # if pass a name_list arg, it will plot all given layer's comparison
    ann_activation, snn_activation = nested_network.plot_comparison(
        nested_input_tensor, spiking_layers_names
    )
    assert len(ann_activation) == len(spiking_layers_names)
    assert len(snn_activation) == len(spiking_layers_names)


def test_deepcopy():
    from copy import deepcopy

    network_copy = deepcopy(network)
    # Make sure copying maintained all parameters and state variables
    compare_networks(network, network_copy)


def test_reset_states():
    from copy import deepcopy

    from sinabs.layers import StatefulLayer

    mynet = deepcopy(nested_network)
    # Perform a forward pass to initialize states
    nested_input_tensor = torch.rand([1, 2, 128, 128])
    with torch.no_grad():
        mynet.spiking_model(nested_input_tensor)
    # Reset network to zeros
    mynet.reset_states(randomize=False)
    for mod in mynet.modules():
        if isinstance(mod, StatefulLayer):
            assert mod.v_mem.any() == False

    # Perform a forward pass to initialize states
    nested_input_tensor = torch.rand([1, 2, 128, 128])
    with torch.no_grad():
        mynet.spiking_model(nested_input_tensor)
    # Reset network to a given range
    value_ranges = [
        {"v_mem": (-4, -2)} for mod in mynet.modules() if isinstance(mod, StatefulLayer)
    ]
    mynet.reset_states(randomize=True, value_ranges=value_ranges)
    for layer in mynet.modules():
        if isinstance(layer, StatefulLayer):
            print(layer, layer.v_mem.shape)

            assert layer.v_mem.max() <= -2
            assert layer.v_mem.min() >= -4



"""
===================
PeriodicExponential
===================
:class:`~sinabs.activation.PeriodicExponential` surrogate gradient.
"""

import matplotlib.pyplot as plt
import torch

import sinabs.activation as sina

x = torch.linspace(-2, 4, 500)
plt.plot(
    x,
    sina.PeriodicExponential()(v_mem=x, spike_threshold=1.0),
    label="PeriodicExponential",
)
plt.xlabel("Neuron membrane potential")
plt.ylabel("Derivative")
plt.legend()



from typing import Callable, Optional, Tuple, Type
from warnings import warn

import torch
from torch import nn

import sinabs
import sinabs.layers as sl
from sinabs.activation import MembraneSubtract, MultiSpike, SingleExponential
from sinabs.conversion import replace_module


def from_model(
    model: nn.Module,
    input_shape: Optional[Tuple[int, int, int]] = None,
    spike_threshold: torch.Tensor = torch.tensor(1.0),
    spike_fn: Callable = MultiSpike,
    reset_fn: Callable = MembraneSubtract(),
    surrogate_grad_fn: Callable = SingleExponential(),
    min_v_mem: float = -1.0,
    bias_rescaling: float = 1.0,
    batch_size: Optional[int] = None,
    num_timesteps: Optional[int] = None,
    synops: bool = False,
    add_spiking_output: bool = False,
    spike_layer_class: Type = sl.IAFSqueeze,
    backend=None,
    kwargs_backend: dict = dict(),
):
    """Converts a Torch model and returns a Sinabs network object. The modules in the model are
    analyzed, and a copy is returned, with all ReLUs and NeuromorphicReLUs turned into
    SpikingLayers.

    Parameters:
        model: Torch model
        input_shape: If provided, the layer dimensions are computed. Otherwise they will be computed at the first forward pass.
        spike_threshold: The membrane potential threshold for spiking (same for all layers).
        spike_fn: The spike dynamics to determine the number of spikes out
        reset_fn: The reset mechanism of the neuron (like reset to zero, or subtract)
        surrogate_grad_fn: The surrogate gradient method for the spiking dynamics
        min_v_mem: The lower bound of the potential in (same for all layers).
        bias_rescaling: Biases are divided by this value.
        batch_size: Must be provided if `num_timesteps` is None and is ignored otherwise.
        num_timesteps: Number of timesteps per sample. If None, `batch_size` must be provided to seperate batch and time dimensions.
        synops: If True (default: False), register hooks for counting synaptic operations during forward passes.
        add_spiking_output: If True (default: False), add a spiking layer to the end of a sequential model if not present.
        spike_layer_class: Can be for example sinabs.layers.IAFSqueeze (default) or EXODUS equivalent.
        backend: String defining the simulation backend (currently sinabs or exodus)
        kwargs_backend: Dict with additional kwargs for the simulation backend
    """
    if backend is not None:
        warn(
            "The 'backend' argument is deprecated and will be removed in a future release, please use spike_layer_class instead."
        )
        _backends = {"sinabs": sl}
        try:
            import sinabs.exodus.layers as el
        except ModuleNotFoundError:
            pass
        else:
            _backends["exodus"] = el
        spike_layer_class = _backends[backend].IAFSqueeze

    try:
        device = next(model.parameters()).device
    except StopIteration:
        device = torch.device("cpu")

    def mapper_fn(module):
        return spike_layer_class(
            spike_threshold=spike_threshold,
            spike_fn=spike_fn,
            reset_fn=reset_fn,
            surrogate_grad_fn=surrogate_grad_fn,
            min_v_mem=min_v_mem,
            batch_size=batch_size,
            num_timesteps=num_timesteps,
            **kwargs_backend,
        ).to(device)

    snn = replace_module(model=model, source_class=nn.ReLU, mapper_fn=mapper_fn)
    snn = replace_module(
        model=snn, source_class=sl.NeuromorphicReLU, mapper_fn=mapper_fn
    )

    if add_spiking_output:
        if isinstance(model, nn.Sequential) and not isinstance(
            model[-1], (nn.ReLU, sl.NeuromorphicReLU)
        ):
            snn.add_module(
                "spike_output",
                spike_layer_class(
                    spike_threshold=spike_threshold,
                    spike_fn=spike_fn,
                    reset_fn=reset_fn,
                    surrogate_grad_fn=surrogate_grad_fn,
                    min_v_mem=min_v_mem,
                    batch_size=batch_size,
                    num_timesteps=num_timesteps,
                    **kwargs_backend,
                ).to(device),
            )
        else:
            warn(
                "Spiking output can only be added to sequential models that do not end in a ReLU. No layer has been added."
            )

    for module in snn.modules():
        if bias_rescaling != 1.0 and isinstance(module, (nn.Linear, nn.Conv2d)):
            if hasattr(module, "bias") and module.bias is not None:
                with torch.no_grad():
                    module.bias.data /= bias_rescaling

    network = sinabs.network.Network(
        model,
        snn,
        input_shape=input_shape,
        synops=synops,
        batch_size=batch_size,
        num_timesteps=num_timesteps,
    )
    return network



"""
========================================
Adaptive Leaky Integrate and Fire (ALIF)
========================================
:class:`~sinabs.layers.ALIF` layer with an adaptive threshold based on the output spikes. 
"""

import torch
from utils import plot_evolution

import sinabs.layers as sl

const_current = torch.ones((1, 100, 1)) * 0.12

alif_neuron = sl.ALIF(
    tau_mem=40.0, tau_adapt=40.0, adapt_scale=20, norm_input=False, record_states=True
)
plot_evolution(alif_neuron, const_current)



import pytest
import torch.nn as nn

import sinabs
import sinabs.layers as sl


def test_layer_replacement_sequential():
    model = nn.Sequential(
        nn.Conv2d(2, 8, 3),
        sl.IAF(spike_threshold=2.0),
        nn.Conv2d(8, 16, 3),
        sl.IAF(spike_threshold=1.0),
        nn.Flatten(),
        nn.Linear(512, 10),
    )

    mapper_fn = lambda module: sl.IAFSqueeze(**module._param_dict, batch_size=18)
    filtered_model = sinabs.conversion.replace_module(
        model, sl.IAF, mapper_fn=mapper_fn
    )

    assert len(model) == len(filtered_model)
    assert type(filtered_model[1]) == sl.IAFSqueeze
    assert filtered_model[1].batch_size == 18
    assert filtered_model[1].spike_threshold == 2
    assert filtered_model[3].spike_threshold == 1
    assert model is not filtered_model


def test_layer_replacement_arbitrary():
    class MyModel(nn.Module):
        def __init__(self):
            super().__init__()
            self.conv = nn.Conv2d(2, 8, 3)
            self.spike1 = sl.IAF(spike_threshold=2.0)
            self.conv2 = nn.Conv2d(8, 16, 3)
            self.spike2 = sl.IAF(spike_threshold=1.0)

    model = MyModel()
    mapper_fn = lambda module: sl.IAFSqueeze(**module._param_dict, batch_size=18)
    filtered_model = sinabs.conversion.replace_module(
        model, sl.IAF, mapper_fn=mapper_fn
    )

    assert type(filtered_model.spike1) == sl.IAFSqueeze
    assert filtered_model.spike1.batch_size == 18
    assert filtered_model.spike1.spike_threshold == 2
    assert filtered_model.spike2.spike_threshold == 1
    assert model is not filtered_model


def test_layer_replacement_individual():
    layer = sl.IAF(spike_threshold=2.0)
    mapper_fn = lambda module: sl.IAFSqueeze(**module.arg_dict, batch_size=4)
    squeezed_layer = sinabs.conversion.replace_module(layer, sl.IAF, mapper_fn)

    assert type(squeezed_layer) == sl.IAFSqueeze
    assert squeezed_layer.batch_size == 4
    assert squeezed_layer.spike_threshold == 2


def test_layer_replacement_individual_in_place():
    layer = sl.IAF(spike_threshold=2.0)
    mapper_fn = lambda module: sl.IAFSqueeze(**module.arg_dict, batch_size=4)
    with pytest.warns(UserWarning):
        squeezed_layer = sinabs.conversion.replace_module_(layer, sl.IAF, mapper_fn)

    assert type(layer) == sl.IAF
    assert squeezed_layer is None



from pbr.version import VersionInfo

__version__ = VersionInfo("sinabs").release_string()

from . import conversion, utils
from .from_torch import from_model
from .network import Network
from .synopcounter import SNNAnalyzer, SynOpCounter
from .utils import reset_states, zero_grad



import pytest
import torch

from sinabs.activation import (
    MaxSpike,
    MembraneReset,
    MembraneSubtract,
    MultiSpike,
    SingleExponential,
    SingleSpike,
)


@pytest.mark.parametrize(
    "spike_fn, reset_fn, output, v_mem",
    [
        (
            SingleSpike,
            MembraneSubtract(),
            torch.tensor([1.0, 0.0]),
            torch.tensor([1.5, 0.3]),
        ),
        (
            SingleSpike,
            MembraneReset(),
            torch.tensor([1.0, 0.0]),
            torch.tensor([0.0, 0.3]),
        ),
        (
            MultiSpike,
            MembraneSubtract(),
            torch.tensor([2.0, 0.0]),
            torch.tensor([0.5, 0.3]),
        ),
        (
            MultiSpike,
            MembraneReset(),
            torch.tensor([2.0, 0.0]),
            torch.tensor([0.0, 0.3]),
        ),
    ],
)
def test_activation_functions(spike_fn, reset_fn, output, v_mem):
    state = {"v_mem": torch.tensor([2.5, 0.3], requires_grad=True)}

    spike_threshold = 1.0

    input_tensors = [state[name] for name in spike_fn.required_states]
    spikes = spike_fn.apply(*input_tensors, spike_threshold, SingleExponential())
    new_state = reset_fn(spikes, state, spike_threshold)

    assert torch.allclose(spikes, output)
    assert torch.allclose(new_state["v_mem"], v_mem)

    loss = torch.nn.functional.mse_loss(spikes, torch.ones_like(spikes))
    loss.backward()

    assert not (state["v_mem"].grad == 0).all()
    assert not torch.isnan(state["v_mem"].grad).any()
    assert not torch.isinf(state["v_mem"].grad).any()


def test_maxspike_activation():
    spike_threshold = 1.0
    max_num_spikes = 4
    spike_fn = MaxSpike(max_num_spikes)
    reset_fn = MembraneSubtract()

    v_mem = torch.tensor([6.5, 1.3], requires_grad=True)
    state = {"v_mem": v_mem}
    spikes = spike_fn.apply(v_mem, spike_threshold, SingleExponential())
    new_state = reset_fn(spikes, state, spike_threshold)

    assert torch.allclose(spikes, torch.tensor([4., 1.]))
    assert torch.allclose(new_state["v_mem"], torch.tensor([2.5, 0.3]))

    loss = torch.nn.functional.mse_loss(spikes, torch.ones_like(spikes))
    loss.backward()

    assert not (state["v_mem"].grad == 0).all()
    assert not torch.isnan(state["v_mem"].grad).any()
    assert not torch.isinf(state["v_mem"].grad).any()


def test_periodic_exponential():
    from sinabs.activation import PeriodicExponential

    grad_fn = PeriodicExponential(grad_width=0.1, grad_scale=1.0)
    x = torch.arange(-5.0, 10.5, 0.01)
    # Must have 10 peaks
    assert torch.sum(grad_fn(x, 1.0) == 1) == 10



from typing import Tuple

import torch
from torch import nn


class Img2SpikeLayer(nn.Module):
    """Layer to convert images to spikes.

    Parameters:
        image_shape: tuple image shape
        tw: int Time window length
        max_rate: maximum firing rate of neurons
        layer_name: string layer name
        norm: the supposed maximum value of the input (default 255.0)
        squeeze: whether to remove singleton dimensions from the input
        negative_spikes: whether to allow negative spikes in response
                         to negative input
    """

    def __init__(
        self,
        image_shape,
        tw: int = 100,
        max_rate: float = 1000,
        norm: float = 255.0,
        squeeze: bool = False,
        negative_spikes: bool = False,
    ):
        super().__init__()
        self.tw = tw
        self.max_rate = max_rate
        self.norm = norm
        self.squeeze = squeeze
        self.negative_spikes = negative_spikes

    def forward(self, img_input):
        if self.squeeze:
            img_input = img_input.squeeze()
        random_tensor = torch.rand(self.tw, *tuple(img_input.shape)).to(
            img_input.device
        )
        if not self.negative_spikes:
            firing_probs = (img_input / self.norm) * (self.max_rate / 1000)
            spk_img = (random_tensor < firing_probs).float()
        else:
            firing_probs = (img_input.abs() / self.norm) * (self.max_rate / 1000)
            spk_img = (random_tensor < firing_probs).float() * img_input.sign().float()
        self.spikes_number = spk_img.abs().sum()
        self.tw = len(spk_img)
        return spk_img

    def get_output_shape(self, input_shape: Tuple):
        # The time dimension is not included in the shape
        # NOTE: This is not true if the squeeze is false but input_shape has a batch_size
        # TODO: Fix this
        return input_shape  # (self.tw, *input_shape)


class Sig2SpikeLayer(torch.nn.Module):
    """Layer to convert analog Signals to spikes.

    Parameters:
        channels_in: number of channels in the analog signal
        tw: int number of time steps for each sample of the signal (up sampling)
        layer_name: string layer name
    """

    def __init__(
        self,
        channels_in,
        tw: int = 1,
        norm_level: float = 1,
        spk_out: bool = True,
    ):
        super().__init__()
        self.tw = tw
        self.norm_level = norm_level
        self.spk_out = spk_out

    def get_output_shape(self, input_shape: Tuple):
        channels, time_steps = input_shape
        return (self.tw * time_steps, channels)

    def forward(self, signal):
        channels, time_steps = signal.shape
        if self.tw != 1:
            signal = signal.view(-1, 1).repeat(1, self.tw).view(channels, -1)
        signal = signal.transpose(1, 0)
        if self.spk_out:
            random_tensor = (
                torch.rand(self.tw * time_steps, channels).to(signal.device)
                * self.norm_level
            )
            spk_sig = (random_tensor < signal).float()
        else:
            # If there is no conversion to spikes
            # just replicate the signal as current injection
            spk_sig = signal

        self.spikes_number = spk_sig.abs().sum()
        return spk_sig



import numpy as np
import pytest
import torch
from torch import nn

import sinabs.layers as sl
from sinabs.activation import MembraneReset, SingleSpike
from sinabs.from_torch import from_model


def test_reconstruct_image():
    # generate random image
    img_shape = (3, 20, 20)
    image = 255.0 * np.random.random(size=img_shape)

    # instantiate layer
    spklayer = sl.Img2SpikeLayer(
        image_shape=img_shape, tw=10000, max_rate=1000.0, squeeze=True
    )

    spikes = spklayer(torch.Tensor(image))
    rates = spikes.mean(0).unsqueeze(0)

    # accept errors of 0.025 for numbers in (0, 1), over 10000 tsteps
    assert np.allclose(rates, image / 255.0, atol=0.025)


def test_reconstruct_real_numbers():
    # generate random image
    input_shape = (3, 20, 20)
    input_data = 2 * np.random.random(size=input_shape) - 1

    # instantiate layer
    spklayer = sl.Img2SpikeLayer(
        image_shape=input_shape,
        tw=10000,
        max_rate=1000.0,
        squeeze=True,
        negative_spikes=True,
        norm=1.0,
    )

    spikes = spklayer(torch.Tensor(input_data))
    rates = spikes.mean(0).unsqueeze(0)

    # accept errors of 0.025 for numbers in (0, 1), over 10000 tsteps
    assert np.allclose(rates, input_data, atol=0.025)


def test_network_conversion_basic():
    class CNN(nn.Module):
        def __init__(self):
            super(CNN, self).__init__()
            self.sequence = nn.Sequential(
                nn.Conv2d(1, 16, kernel_size=(3, 3), bias=False),
                nn.BatchNorm2d(16, affine=True),
                nn.ReLU(),
                nn.AvgPool2d(kernel_size=2),
                nn.Conv2d(16, 8, kernel_size=(3, 3), bias=True),
                nn.BatchNorm2d(8, affine=False, track_running_stats=True),
            )

        def forward(self, x):
            return self.sequence(x)

    input_shape = (1, 28, 28)

    cnn = CNN().eval()

    cnn.sequence[1].running_mean = 2.0 * torch.rand(16) - 1.0
    cnn.sequence[1].running_var = torch.rand(16) + 1.0
    cnn.sequence[1].weight = nn.Parameter(2 * torch.rand(16) - 1.0)
    cnn.sequence[1].bias = nn.Parameter(2 * torch.rand(16) - 1.0)

    cnn.sequence[5].running_mean = 2.0 * torch.rand(8) - 1.0
    cnn.sequence[5].running_var = torch.rand(8) + 1.0

    img2spk = sl.Img2SpikeLayer(image_shape=input_shape, tw=1000, norm=1.0)
    snn = from_model(cnn, input_shape=input_shape, batch_size=1)

    img = torch.Tensor(np.random.random(size=input_shape))

    with torch.no_grad():
        spk_img = img2spk(img)
        snn_res = snn(spk_img).mean(0)
        cnn_res = cnn(img.unsqueeze(0))

    # import matplotlib.pyplot as plt
    # plt.plot(snn_res.numpy().ravel(), cnn_res.numpy().ravel(), '.')
    # plt.show()

    assert np.allclose(snn_res, cnn_res, atol=0.025)


def test_network_conversion_add_spk_out():
    cnn = nn.Sequential(
        nn.Conv2d(1, 16, kernel_size=(3, 3), bias=False),
        nn.BatchNorm2d(16, affine=True),
        nn.ReLU(),
        nn.AvgPool2d(kernel_size=2),
        nn.Conv2d(16, 8, kernel_size=(3, 3), bias=True),
        nn.BatchNorm2d(8, affine=False, track_running_stats=True),
    )

    input_shape = (1, 28, 28)

    cnn[1].running_mean = 2.0 * torch.rand(16) - 1.0
    cnn[1].running_var = torch.rand(16) + 1.0
    cnn[1].weight = nn.Parameter(2 * torch.rand(16) - 1.0)
    cnn[1].bias = nn.Parameter(2 * torch.rand(16) - 1.0)

    cnn[5].running_mean = 2.0 * torch.rand(8) - 1.0
    cnn[5].running_var = torch.rand(8) + 1.0

    img2spk = sl.Img2SpikeLayer(image_shape=input_shape, tw=1000, norm=1.0)

    snn = from_model(
        cnn, input_shape=input_shape, add_spiking_output=True, batch_size=1
    )

    mod_names = [name for name, mod in cnn.named_modules()]
    assert "spike_output" not in mod_names
    assert isinstance(snn.spiking_model.spike_output, sl.StatefulLayer)

    img = torch.Tensor(np.random.random(size=input_shape))

    with torch.no_grad():
        spk_img = img2spk(img)
        snn(spk_img).mean(0)


def test_network_conversion_complicated_model():
    """Try converting rather complicated network model with nested structures, which used to fail
    before."""

    ann = nn.Sequential(
        nn.Conv2d(1, 1, 1),
        nn.Sequential(nn.AvgPool2d(2), nn.ReLU(), nn.Conv2d(1, 1, 1)),
        nn.ReLU(),
        nn.AvgPool2d(3),
        nn.Sequential(
            nn.Conv2d(1, 1, 2), nn.Sequential(nn.Flatten(), nn.Linear(400, 12))
        ),
    )

    snn = from_model(ann, batch_size=1)


def test_network_conversion_with_batch_size():
    ann = nn.Sequential(
        nn.Conv2d(1, 1, 1),
        nn.Sequential(nn.AvgPool2d(2), nn.ReLU(), nn.Conv2d(1, 1, 1)),
        nn.ReLU(),
        nn.AvgPool2d(3),
        nn.Sequential(
            nn.Conv2d(1, 1, 2), nn.Sequential(nn.Flatten(), nn.Linear(400, 12))
        ),
    )

    # Model conversion without input size specification
    snn = from_model(ann, batch_size=32)
    assert snn.spiking_model[2].batch_size == 32

    # Model conversion with input size specification
    snn = from_model(ann, input_shape=(1, 128, 128), batch_size=32)
    assert snn.spiking_model[2].batch_size == 32


def test_network_conversion_with_num_timesteps():
    ann = nn.Sequential(
        nn.Conv2d(1, 1, 1),
        nn.Sequential(nn.AvgPool2d(2), nn.ReLU(), nn.Conv2d(1, 1, 1)),
        nn.ReLU(),
        nn.AvgPool2d(3),
        nn.Sequential(
            nn.Conv2d(1, 1, 2), nn.Sequential(nn.Flatten(), nn.Linear(400, 12))
        ),
    )

    # Model conversion without input size specification
    snn = from_model(ann, num_timesteps=32)
    assert snn.spiking_model[2].num_timesteps == 32

    # Model conversion with input size specification
    snn = from_model(ann, input_shape=(1, 128, 128), num_timesteps=32)
    assert snn.spiking_model[2].num_timesteps == 32


def test_network_conversion_backend():
    """Try conversion with sinabs explicitly stated as backend."""

    ann = nn.Sequential(
        nn.Conv2d(1, 1, 1),
        nn.Sequential(nn.AvgPool2d(2), nn.ReLU(), nn.Conv2d(1, 1, 1)),
        nn.ReLU(),
        nn.AvgPool2d(3),
        nn.Sequential(
            nn.Conv2d(1, 1, 2), nn.Sequential(nn.Flatten(), nn.Linear(400, 12))
        ),
    )

    with pytest.warns(UserWarning):
        snn = from_model(ann, backend="sinabs", batch_size=1)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_parameter_devices():
    cnn = nn.Sequential(nn.Conv2d(1, 16, kernel_size=(3, 3), bias=False), nn.ReLU())

    batch_size = 10
    snn = from_model(cnn.cuda(), input_shape=None, batch_size=batch_size)

    spk_img = torch.rand((batch_size, 1, 28, 28)).cuda()

    with torch.no_grad():
        snn(spk_img)


def test_activation_change():
    ann = nn.Sequential(
        nn.Conv2d(1, 4, (3, 3), padding=1),
        nn.ReLU(),
        sl.SumPool2d(2, 2),
        nn.Conv2d(1, 4, (3, 3), padding=1),
        nn.ReLU(),
    )

    network = from_model(
        ann,
        spike_threshold=3.0,
        spike_fn=SingleSpike,
        reset_fn=MembraneReset(),
        min_v_mem=-3.0,
        batch_size=1,
    )

    spk_layer: sl.IAFSqueeze = network.spiking_model[1]

    # Test number of spikes generated against neuron threshold
    input_data = torch.zeros(10, 4, 32, 32)
    input_data[:, 0, 0, 0] = 1.0
    out = spk_layer(input_data)
    assert out.sum() == 10 // 3

    # Test  input much larger than threshold
    spk_layer.reset_states()
    input_data = torch.zeros(10, 4, 32, 32)
    input_data[0, 0, 0, 0] = 7
    out = spk_layer(input_data)
    assert out.sum() == 1

    # Test same for the second layer
    spk_layer: sl.IAFSqueeze = network.spiking_model[4]

    # Test number of spikes generated against neuron threshold
    input_data = torch.zeros(10, 4, 32, 32)
    input_data[:, 0, 0, 0] = 1.0
    out = spk_layer(input_data)
    assert out.sum() == 10 // 3

    # Test  input much larger than threshold
    spk_layer.reset_states()
    input_data = torch.zeros(10, 4, 32, 32)
    input_data[0, 0, 0, 0] = 7
    out = spk_layer(input_data)
    assert out.sum() == 1



from typing import Callable, Optional

import torch
import torch.nn as nn


class Repeat(nn.Module):
    """Utility layer which wraps any nn.Module.

    It flattens time and batch dimensions of the input before feeding it to the child module and
    unflattens those dimensions to the original shape before passing it to the next layer.
    """

    def __init__(self, module: nn.Module):
        super().__init__()
        self.module = module

    def forward(self, x):
        orig_shape = x.shape[:2]
        x = x.flatten(0, 1)
        x = self.module(x)
        return x.unflatten(0, orig_shape)

    def __repr__(self):
        return "Repeated " + self.module.__repr__()


class FlattenTime(nn.Flatten):
    """Utility layer which always flattens first two dimensions and is a special case of
    `torch.nn.Flatten()`. Meant to convert a tensor of dimensions (Batch, Time, Channels, Height,
    Width) into a tensor of (Batch*Time, Channels, Height, Width).

    Shape:
        - Input: :math:`(Batch, Time, Channel, Height, Width)` or :math:`(Batch, Time, Channel)`
        - Output: :math:`(Batch \\times Time, Channel, Height, Width)` or :math:`(Batch \\times Time, Channel)`
    """

    def __init__(self):
        super().__init__(start_dim=0, end_dim=1)


class UnflattenTime(nn.Module):
    """Utility layer which always unflattens (expands) the first dimension into two separate ones.
    Meant to convert a tensor of dimensions (Batch*Time, Channels, Height, Width) into a tensor of
    (Batch, Time, Channels, Height, Width).

    Shape:
        - Input: :math:`(Batch \\times Time, Channel, Height, Width)` or :math:`(Batch \\times Time, Channel)`
        - Output: :math:`(Batch, Time, Channel, Height, Width)` or :math:`(Batch, Time, Channel)`
    """

    def __init__(self, batch_size: int):
        super().__init__()
        self.batch_size = batch_size

    def forward(self, x):
        num_time_steps = x.shape[0] // self.batch_size
        return x.unflatten(0, (self.batch_size, num_time_steps))


class SqueezeMixin:
    """Utility mixin class that will wrap the __init__ and forward call to flatten the input to and
    the output from a child class.

    The wrapped __init__ will provide two additional parameters batch_size and num_timesteps and
    the wrapped forward will unpack and repack the first dimension into batch and time.
    """

    def squeeze_init(self, batch_size: Optional[int], num_timesteps: Optional[int]):
        if not batch_size and not num_timesteps:
            raise TypeError("You need to specify either batch_size or num_timesteps.")
        if not batch_size:
            batch_size = -1
        if not num_timesteps:
            num_timesteps = -1
        self.batch_size = int(batch_size)
        self.num_timesteps = int(num_timesteps)

    def squeeze_forward(self, input_data: torch.Tensor, forward_method: Callable):
        inflated_input = input_data.reshape(
            self.batch_size, self.num_timesteps, *input_data.shape[1:]
        )
        inflated_output = forward_method(inflated_input)
        return inflated_output.flatten(start_dim=0, end_dim=1)

    def squeeze_param_dict(self, param_dict: dict) -> dict:
        param_dict.update(
            batch_size=self.batch_size,
            num_timesteps=self.num_timesteps,
        )
        return param_dict



from typing import Tuple


def conv_output_size(image_length: int, kernel_length: int, stride: int) -> int:
    """Computes output dimension given input dimension, kernel size and stride, assumign no
    padding, *per* dimension given.

    :param image_length: int image size on one dimension
    :param kernel_length: int kernel_length size on one dimension
    :param stride: int Stride size on one dimension
    :return: int -- convolved output image size on one dimension
    """
    try:
        assert image_length >= kernel_length
    except AssertionError:
        raise Exception(
            "Image dimension {0} smaller than kernel dimension {1}".format(
                image_length, kernel_length
            )
        )
    return int((image_length - kernel_length) / stride + 1)


def compute_same_padding_size(kernel_length: int) -> (int, int):
    """Computes padding for 'same' padding *per* dimension given.

    :param kernel_length: int Kernel size
    :returns: Tuple -- (padStart, padStop) , padding on left/right or top/bottom

    Note : Only works for stride 1
    """
    start = kernel_length // 2
    stop = start - 1 if kernel_length % 2 == 0 else start
    return start, stop


def compute_padding(
    kernel_shape: tuple, input_shape: tuple, mode="valid"
) -> (int, int, int, int):
    """Computes padding for 'same' or 'valid' padding.

    :param kernel_shape: Kernel shape (height, width)
    :param input_shape: Input shape (channels, height, width)
    :param mode: `valid` or `same`
    :return: Tuple -- (pad_left, pad_right, pad_top, pad_bottom)
    """
    if mode == "valid":
        padding = (0, 0, 0, 0)
    elif mode == "same":
        padding = (
            *compute_same_padding_size(kernel_shape[1]),
            *compute_same_padding_size(kernel_shape[0]),
        )
    else:
        raise NotImplementedError("Unknown padding mode")
    return padding


def infer_output_shape(torch_layer, input_shape: Tuple) -> Tuple:
    """Compute the output dimensions given input dimensions.

    :param torch_layer: a Torch layer
    :param input_shape: the shape of the input tensor
    :return: Tuple -- the size of output tensor
    """
    import torch

    with torch.no_grad():
        # Gen random input
        tsr = torch.rand((1, *input_shape))
        tsrOut = torch_layer(tsr)
        output_shape = tuple(tsrOut.shape)[1:]

    return output_shape



import matplotlib.pyplot as plt
import torch

import sinabs
import sinabs.layers as sl


def plot_evolution(neuron_model: sinabs.layers, input: torch.Tensor):
    neuron_model.reset_states()
    output = neuron_model(input)

    plt.figure(figsize=(10, 3))
    for key, recording in neuron_model.recordings.items():
        plt.plot(recording.detach().flatten(), drawstyle="steps", label=key)
    plt.plot(
        output.detach().flatten(), drawstyle="steps", color="black", label="output"
    )
    if not "spike_threshold" in neuron_model.recordings:
        plt.plot(
            [neuron_model.spike_threshold] * input.shape[1], label="spike_threshold"
        )
    plt.xlabel("time")
    plt.title(f"{neuron_model.__class__.__name__} neuron dynamics")
    plt.legend()



import warnings
from functools import partial

import torch
from numpy import product

import sinabs.layers as sl
from sinabs.layers import NeuromorphicReLU


def spiking_hook(self, input_, output):
    """Forward hook meant for Sinabs/Exodus layers.

    Calculates n_neurons (scalar), firing_rate_per_neuron (C,H,W) and average firing_rate (scalar).
    """
    if isinstance(self, sl.SqueezeMixin):
        output = output.reshape(self.batch_size, self.num_timesteps, *output.shape[1:])
    self.n_neurons = output[0, 0].numel()
    if self.firing_rate_per_neuron == None:
        self.firing_rate_per_neuron = output.mean((0, 1))
    else:
        self.firing_rate_per_neuron = self.firing_rate_per_neuron + output.mean((0, 1))
    self.tracked_firing_rate = self.tracked_firing_rate + output.mean()
    self.n_batches = self.n_batches + 1


def synops_hook(unflattened_shape, self, input_, out):
    """Forward hook for parameter layers such as Conv2d or Linear.

    Calculates the total amount of input and output and the number of synaptic operations. The
    `unflattened_shape` parameter is a bit of a hack since Sinabs sometimes uses squeezed inputs
    and from inside the hook we cannot know whether the input is squeezed or not.
    """
    assert len(input_) == 1, "Multiple inputs not supported for synops hook"
    input_ = input_[0]
    if unflattened_shape is not None:
        batch_size, num_timesteps = unflattened_shape
        input_ = input_.reshape(batch_size, num_timesteps, *input_.shape[1:])
    self.synops = self.synops + input_.sum() * self.fanout
    self.n_samples = self.n_samples + input_.shape[0]
    self.num_timesteps = input_.shape[1]


class SNNAnalyzer:
    """Helper class to acquire statistics for spiking and parameter layers at the same time.

    Parameters:
        model: Your PyTorch model.
        dt: the number of milliseconds corresponding to a time step in the simulation (default 1.0).

    Example:
        >>> analyzer = SNNAnalyzer(my_spiking_model)
        >>> output = my_spiking_model(input_)  # forward pass
        >>> layer_stats = analyzer.get_layer_statistics()
        >>> model_stats = analyzer.get_model_statistics()
    """

    def __init__(self, model: torch.nn.Module, dt: float = 1.0):
        self.model = model
        self.dt = dt
        self.handles = []
        self._setup_hooks()

    def _setup_hooks(self):
        """Attaches spiking and parameter hooks to matching layers and resets all counters."""
        # This is a hack to unflatten inputs/outputs in conv/linear layers
        # inside the hook since we cannot know from the shape alone whether
        # it is flattened (B*T,C,H,W) or unflattened (B,T,C,H,W)
        unflattened_shape = None
        for layer in self.model.modules():
            if isinstance(layer, sl.SqueezeMixin):
                unflattened_shape = (layer.batch_size, layer.num_timesteps)

        for layer in self.model.modules():
            if isinstance(layer, sl.StatefulLayer):
                layer.firing_rate_per_neuron = None
                layer.tracked_firing_rate = 0
                layer.n_batches = 0
                handle = layer.register_forward_hook(spiking_hook)
                self.handles.append(handle)
            if isinstance(layer, torch.nn.Conv2d):
                layer.fanout = (
                    layer.out_channels
                    * product(layer.kernel_size)
                    / product(layer.stride)
                )
                layer.synops = 0
                layer.n_samples = 0
                handle = layer.register_forward_hook(
                    partial(synops_hook, unflattened_shape)
                )
                self.handles.append(handle)
            elif isinstance(layer, torch.nn.Linear):
                layer.fanout = layer.out_features
                layer.synops = 0
                layer.n_samples = 0
                handle = layer.register_forward_hook(
                    partial(synops_hook, unflattened_shape)
                )
                self.handles.append(handle)

    def __del__(self):
        for handle in self.handles:
            handle.remove()

    def get_layer_statistics(self) -> dict:
        spike_dict = {}
        spike_dict["spiking"] = {}
        spike_dict["parameter"] = {}
        scale_facts = []
        for name, module in self.model.named_modules():
            if (
                hasattr(module, "tracked_firing_rate")
                or hasattr(module, "firing_rate_per_neuron")
                or hasattr(module, "n_neurons")
            ):
                spike_dict["spiking"][name] = {}
            if hasattr(module, "tracked_firing_rate"):
                spike_dict["spiking"][name].update(
                    {"firing_rate": module.tracked_firing_rate / module.n_batches}
                )
            if hasattr(module, "firing_rate_per_neuron"):
                spike_dict["spiking"][name].update(
                    {
                        "firing_rate_per_neuron": module.firing_rate_per_neuron
                        / module.n_batches
                    }
                )
            if hasattr(module, "n_neurons"):
                spike_dict["spiking"][name].update({"n_neurons": module.n_neurons})

            # synops statistics
            if isinstance(module, torch.nn.AvgPool2d):
                if module.kernel_size != module.stride:
                    warnings.warn(
                        f"In order for the Synops counter to work accurately the pooling "
                        f"layers kernel size should match their strides. At the moment at layer {name}, "
                        f"the kernel_size = {module.kernel_size}, the stride = {module.stride}."
                    )
                ks = module.kernel_size
                scale_factor = ks**2 if isinstance(ks, int) else ks[0] * ks[1]
                scale_facts.append(scale_factor)
            if hasattr(module, "synops"):
                scale_factor = 1
                while len(scale_facts) != 0:
                    scale_factor *= scale_facts.pop()
                spike_dict["parameter"][name] = {
                    "fanout_prev": module.fanout,
                    "synops": module.synops / module.n_samples * scale_factor,
                    "num_timesteps": module.num_timesteps,
                    "time_window": module.num_timesteps * self.dt,
                    "SynOps/s": (module.synops / module.n_samples * scale_factor)
                    / module.num_timesteps
                    / self.dt
                    * 1000,
                }
        return spike_dict

    def get_model_statistics(self):
        stats_dict = {}
        firing_rates = []
        synops = 0.0
        for name, module in self.model.named_modules():
            if hasattr(module, "firing_rate_per_neuron"):
                firing_rates.append(
                    module.firing_rate_per_neuron.ravel() / module.n_batches
                )
            if hasattr(module, "synops"):
                synops = synops + module.synops / module.n_samples
        if len(firing_rates) > 0:
            stats_dict["firing_rate"] = torch.cat(firing_rates).mean()
        stats_dict["synops"] = synops
        return stats_dict

    def reset(self):
        for handle in self.handles:
            handle.remove()
        self._setup_hooks()


class SynOpCounter:
    """Counter for the synaptic operations emitted by all Neuromorphic ReLUs in an analog CNN
    model.

    Parameters:
        modules: list of modules, e.g. MyTorchModel.modules()
        sum_activations: If True (default), returns a single number of synops, otherwise a list of layer synops.

    Example:
        >>> counter = SynOpCounter(MyTorchModel.modules(), sum_activations=True)
        >>> output = MyTorchModule(input)  # forward pass
        >>> synop_count = counter()
    """

    def __init__(self, modules, sum_activations=True):
        self.modules = []
        for module in modules:
            if isinstance(module, NeuromorphicReLU) and module.fanout > 0:
                self.modules.append(module)

        if len(self.modules) == 0:
            raise ValueError("No NeuromorphicReLU found in module list.")

        self.sum_activations = sum_activations

    def __call__(self):
        synops = []
        for module in self.modules:
            synops.append(module.activity)

        if self.sum_activations:
            synops = torch.stack(synops).sum()
        return synops



def test_forward():
    import torch

    from sinabs.layers import Sig2SpikeLayer

    channels = 4
    tw = 5

    lyr = Sig2SpikeLayer(
        channels_in=channels,
        tw=tw,
    )

    sig = torch.tensor([[1.0, 0.5, 0.25]] * channels)
    spk = lyr(sig)
    assert spk.shape == (tw * 3, channels)



def test_quantize():
    import torch

    from sinabs.layers import NeuromorphicReLU

    x = torch.rand(20, requires_grad=True)
    lyr = NeuromorphicReLU(quantize=True, fanout=1, stochastic_rounding=False)

    interm = lyr(x)
    out = 40 * interm
    err = (out.sum() - 100) ** 2
    err.backward()
    assert out.sum() == 0
    assert x.grad.sum() != 0


def test_stochastic_rounding():
    import torch

    from sinabs.layers import NeuromorphicReLU

    x = torch.rand(20, requires_grad=True)
    lyr = NeuromorphicReLU(quantize=True, fanout=1, stochastic_rounding=True)

    interm = lyr(x)
    out = 40 * interm
    err = (out.sum() - 100) ** 2
    err.backward()
    assert out.sum() > 0
    assert x.grad.sum() != 0



from itertools import product

import numpy as np
import pytest
import torch
import torch.nn as nn

import sinabs.activation as sa
from sinabs.layers import LIF, LIFRecurrent


def test_lif_basic():
    torch.set_printoptions(precision=10)
    batch_size, time_steps = 10, 100
    tau_mem = torch.tensor(20.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = LIF(tau_mem=tau_mem, min_v_mem=-2.0)
    spike_output = layer(input_current)

    # Make sure __repr__ works
    repr(layer)

    # Make sure arg_dict works
    layer.arg_dict

    assert "min_v_mem" in layer.state_dict().keys()
    assert "spike_threshold" in layer.state_dict().keys()
    assert layer.does_spike
    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_lif_array_tau():
    torch.set_printoptions(precision=10)
    batch_size, time_steps = 10, 100
    # Shape for tau mem such that 2nd dimension needs to be expanded
    tau_mem = torch.randn(2, 1, 7) + 20.0
    tau_syn = torch.randn(2, 1, 7) + 20.0
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = LIF(tau_mem=tau_mem, tau_syn=tau_syn)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_lif_v_mem_recordings():
    batch_size, time_steps = 10, 100
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7)
    input_current[:, :5] = 0
    layer = LIF(tau_mem=20.0, norm_input=False, record_states=True)
    spike_output = layer(input_current)

    layer.recordings["v_mem"].sum().backward()
    assert layer.tau_mem.grad is not None

    assert layer.recordings["v_mem"].shape == spike_output.shape
    # Ensure causality
    assert (layer.recordings["v_mem"][:, :5] == 0).all()
    assert "i_syn" not in layer.recordings.keys()


def test_lif_i_syn_recordings():
    batch_size, time_steps = 10, 100
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7)
    layer = LIF(tau_mem=20.0, tau_syn=10.0, norm_input=False, record_states=True)
    spike_output = layer(input_current)

    layer.recordings["i_syn"].sum().backward()
    assert layer.tau_syn.grad is not None

    assert layer.recordings["v_mem"].shape == spike_output.shape
    assert layer.recordings["i_syn"].shape == spike_output.shape


def test_lif_recurrent_v_mem_recordings():
    batch_size, time_steps, n_neurons = 10, 100, 20
    input_current = torch.rand(batch_size, time_steps, n_neurons)
    rec_connect = nn.Sequential(
        nn.Flatten(), nn.Linear(n_neurons, n_neurons, bias=False)
    )
    layer = LIFRecurrent(
        tau_mem=20.0, rec_connect=rec_connect, norm_input=False, record_states=True
    )
    spike_output = layer(input_current)

    layer.recordings["v_mem"].sum().backward()
    assert layer.tau_mem.grad is not None

    assert layer.recordings["v_mem"].shape == spike_output.shape
    assert "i_syn" not in layer.recordings.keys()


def test_lif_recurrent_i_syn_recordings():
    batch_size, time_steps, n_neurons = 10, 100, 20
    input_current = torch.rand(batch_size, time_steps, n_neurons)
    rec_connect = nn.Sequential(
        nn.Flatten(), nn.Linear(n_neurons, n_neurons, bias=False)
    )
    layer = LIFRecurrent(
        tau_mem=20.0,
        tau_syn=10.0,
        rec_connect=rec_connect,
        norm_input=False,
        record_states=True,
    )
    spike_output = layer(input_current)

    layer.recordings["i_syn"].sum().backward()
    assert layer.tau_syn.grad is not None

    assert layer.recordings["v_mem"].shape == spike_output.shape
    assert layer.recordings["i_syn"].shape == spike_output.shape


def test_lif_single_spike():
    torch.set_printoptions(precision=10)
    batch_size, time_steps = 10, 100
    tau_mem = torch.tensor(20.0)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) * 20
    layer = LIF(tau_mem=tau_mem, spike_fn=sa.SingleSpike)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0
    assert torch.max(spike_output) == 1


def test_lif_max_spike():
    torch.set_printoptions(precision=10)
    batch_size, time_steps = 10, 100
    tau_mem = torch.tensor(20.0)
    max_spikes = 3
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) * 100
    layer = LIF(tau_mem=tau_mem, spike_fn=sa.MaxSpike(max_spikes))
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0
    assert torch.max(spike_output) == max_spikes


def test_lif_with_current_dynamics():
    batch_size, time_steps = 10, 100
    tau_mem = torch.tensor(30.0)
    tau_syn = torch.tensor(10.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = LIF(tau_mem=tau_mem, tau_syn=tau_syn)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_lif_train_alphas():
    batch_size, time_steps = 10, 100
    tau_mem = torch.tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = LIF(tau_mem=tau_mem, train_alphas=True)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_lif_train_alphas_with_current_dynamics():
    batch_size, time_steps = 10, 100
    tau_mem = torch.tensor(30.0)
    tau_syn = torch.tensor(10.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = LIF(tau_mem=tau_mem, tau_syn=tau_syn, train_alphas=True)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() > 0


def test_lif_input_integration():
    batch_size, time_steps = 10, 100
    tau_mem = torch.tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.zeros(batch_size, time_steps, 2, 7, 7)
    # only inject current in the first time step
    input_current[:, 0] = 1 / (1 - alpha)
    layer = LIF(tau_mem=tau_mem)
    spike_output = layer(input_current)

    assert (
        spike_output.sum() == np.product(input_current.shape) / time_steps
    ), "Every neuron should spike exactly once."
    assert (
        spike_output[:, 0].sum() == spike_output.sum()
    ), "First output time step should contain all the spikes."


def test_lif_membrane_decay():
    batch_size, time_steps = 10, 100
    tau_mem = torch.tensor(30.0)
    start_value = 0.75
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.zeros(batch_size, time_steps, 2, 7, 7)
    # only inject current in the first time step
    input_current[:, 0] = start_value / (1 - alpha)
    layer = LIF(tau_mem=tau_mem)
    spike_output = layer(input_current)

    # decay only starts after first time step
    membrane_decay = start_value * alpha ** (time_steps - 1)
    # account for rounding errors with .isclose()
    assert torch.isclose(
        layer.v_mem, membrane_decay, atol=1e-08
    ).all(), "Neuron membrane potentials do not seems to decay correctly."


def test_lif_recurrent_basic():
    batch_size, time_steps = 10, 100
    tau_mem = torch.tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_dimensions = (batch_size, time_steps, 2, 10)
    n_neurons = np.product(input_dimensions[2:])
    input_current = torch.ones(*input_dimensions) * 0.5 / (1 - alpha)

    rec_connect = nn.Sequential(
        nn.Flatten(), nn.Linear(n_neurons, n_neurons, bias=False)
    )
    rec_connect[1].weight = nn.Parameter(
        torch.ones(n_neurons, n_neurons) / n_neurons * 0.5 / (1 - alpha)
    )
    layer = LIFRecurrent(tau_mem=tau_mem, rec_connect=rec_connect)
    spike_output = layer(input_current)

    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() == batch_size * (time_steps - 2) * 2 * 10


def test_lif_with_shape():
    batch_size, time_steps, n_neurons = 10, 100, 5
    tau_mem = torch.tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.zeros((batch_size, time_steps, n_neurons))
    input_current[:, 0] = 1 / (1 - alpha)

    layer = LIF(tau_mem=tau_mem, shape=(batch_size, n_neurons))
    v_mem_shape_intitial = layer.v_mem.shape
    spike_output = layer(input_current)

    assert layer.v_mem.shape == v_mem_shape_intitial
    assert input_current.shape == spike_output.shape
    assert torch.isnan(spike_output).sum() == 0
    assert spike_output.sum() == batch_size * n_neurons


def test_lif_with_multiple_taus():
    batch_size, time_steps, n_neurons = 2, 2, 5
    tau_mem = torch.tensor([i * 10.0 for i in range(1, 6)])
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.zeros((batch_size, time_steps, n_neurons))
    # make sure that v_mem after the first ts is beneath spike threshold
    v_mem = 0.75
    input_current[:, 0] = v_mem / (1 - alpha)

    layer = LIF(tau_mem=tau_mem)
    layer(input_current)

    assert torch.allclose(layer.v_mem[0], alpha * v_mem)
    assert (layer.v_mem[0] == layer.v_mem[1]).all()


def test_lif_firing_rate():
    batch_size, time_steps, n_neurons = 5, 10, 5
    tau_mem = 20.0
    input_current = torch.zeros((batch_size, time_steps, n_neurons))
    input_current[:, 0] = 1

    layer = LIF(tau_mem=tau_mem, norm_input=False)
    spikes = layer(input_current)

    assert layer.firing_rate > 0
    assert layer.firing_rate == spikes.sum() / (batch_size * time_steps * n_neurons)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_lif_on_gpu():
    batch_size, time_steps = 5, 20
    tau_mem = torch.as_tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_current = torch.rand(batch_size, time_steps, 2, 7, 7) / (1 - alpha)
    layer = LIF(tau_mem=tau_mem)

    layer = layer.to("cuda")
    layer(input_current.to("cuda"))


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA not available")
def test_lif_recurrent_on_gpu():
    batch_size, time_steps = 5, 20
    tau_mem = torch.as_tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_dimensions = (batch_size, time_steps, 2, 10)
    n_neurons = np.product(input_dimensions[2:])
    input_current = torch.ones(*input_dimensions) * 0.5 / (1 - alpha)

    rec_connect = nn.Sequential(
        nn.Flatten(), nn.Linear(n_neurons, n_neurons, bias=False)
    )
    layer = LIFRecurrent(tau_mem=tau_mem, rec_connect=rec_connect)

    layer = layer.to("cuda")
    layer(input_current.to("cuda"))


def test_min_v_mem():
    batch_size, time_steps = 10, 1
    tau_mem = torch.tensor(30.0)
    alpha = torch.exp(-1 / tau_mem)
    input_data = torch.rand(batch_size, time_steps, 2, 7, 7) / -(1 - alpha)
    layer = LIF(tau_mem=tau_mem)
    layer(input_data)
    assert (layer.v_mem < -0.5).any()

    layer = LIF(tau_mem=tau_mem, min_v_mem=-0.5)
    layer(input_data)
    assert not (layer.v_mem < -0.5).any()


params = product((20.0, None), (True, False))


@pytest.mark.parametrize("tau_syn,train_alphas", params)
def test_alpha_tau_conversion(tau_syn, train_alphas):
    tau_mem = torch.rand((3, 4)) * 20 + 30
    alpha_mem = torch.exp(-1.0 / tau_mem)

    layer = LIF(tau_mem=tau_mem, tau_syn=tau_syn, train_alphas=train_alphas)
    assert torch.isclose(layer.tau_mem_calculated, tau_mem).all()
    assert torch.isclose(layer.alpha_mem_calculated, alpha_mem).all()
    if tau_syn is None:
        assert layer.tau_syn_calculated is None
        assert layer.alpha_syn_calculated is None
    else:
        tau_syn = torch.as_tensor(tau_syn)
        alpha_syn = torch.exp(-1.0 / tau_syn)
        assert torch.isclose(layer.tau_syn_calculated, tau_syn).all()
        assert torch.isclose(layer.alpha_syn_calculated, alpha_syn).all()
